#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef
from .grhi import hud_background

warning_sound = Struct("warning_sound",
    dependency("sound", ('lsnd', 'snd!')),
    Bool32("latched_to",
        "shield_recharging",
        "shield_recharged",
        "shield_low",
        "shield_empty",
        "health_low",
        "health_empty",
        "health_minor_damage",
        "health_major_damage",
        ),
    Float("scale"),
    SIZE=56
    )

shield_panel_meter = Struct("shield_panel_meter",
    QStruct("anchor_offset",
        SInt16("x"), SInt16("y"), ORIENT='h',
        ),
    Float("width_scale"),
    Float("height_scale"),
    Bool16("scaling_flags", *hud_scaling_flags),

    Pad(22),
    dependency("meter_bitmap", "bitm"),
    #QStruct("color_at_meter_minimum", INCLUDE=xrgb_byte),
    #QStruct("color_at_meter_maximum", INCLUDE=xrgb_byte),
    #QStruct("flash_color", INCLUDE=xrgb_byte),
    #QStruct("empty_color", INCLUDE=argb_byte),
    UInt32("color_at_meter_minimum", INCLUDE=xrgb_uint32),
    UInt32("color_at_meter_maximum", INCLUDE=xrgb_uint32),
    UInt32("flash_color", INCLUDE=xrgb_uint32),
    UInt32("empty_color", INCLUDE=argb_uint32),
    Bool8("flags", *hud_panel_meter_flags),
    SInt8("minimum_meter_value"),
    SInt16("sequence_index"),
    SInt8("alpha_multiplier"),
    SInt8("alpha_bias"),
    SInt16("value_scale"),
    Float("opacity"),
    Float("translucency"),
    #QStruct("disabled_color", INCLUDE=argb_byte),
    UInt32("disabled_color", INCLUDE=argb_uint32),
    Pad(16),
    #QStruct("overcharge_minimum_color", INCLUDE=xrgb_byte),
    #QStruct("overcharge_maximum_color", INCLUDE=xrgb_byte),
    #QStruct("overcharge_flash_color", INCLUDE=xrgb_byte),
    #QStruct("overcharge_empty_color", INCLUDE=xrgb_byte),
    UInt32("overcharge_minimum_color", INCLUDE=xrgb_uint32),
    UInt32("overcharge_maximum_color", INCLUDE=xrgb_uint32),
    UInt32("overcharge_flash_color", INCLUDE=xrgb_uint32),
    UInt32("overcharge_empty_color", INCLUDE=xrgb_uint32),
    SIZE=136
    )

health_panel_meter = Struct("health_panel_meter",
    QStruct("anchor_offset",
        SInt16("x"), SInt16("y"), ORIENT='h',
        ),
    Float("width_scale"),
    Float("height_scale"),
    Bool16("scaling_flags", *hud_scaling_flags),

    Pad(22),
    dependency("meter_bitmap", "bitm"),
    #QStruct("color_at_meter_minimum", INCLUDE=xrgb_byte),
    #QStruct("color_at_meter_maximum", INCLUDE=xrgb_byte),
    #QStruct("flash_color", INCLUDE=xrgb_byte),
    #QStruct("empty_color", INCLUDE=argb_byte),
    UInt32("color_at_meter_minimum", INCLUDE=xrgb_uint32),
    UInt32("color_at_meter_maximum", INCLUDE=xrgb_uint32),
    UInt32("flash_color", INCLUDE=xrgb_uint32),
    UInt32("empty_color", INCLUDE=argb_uint32),
    Bool8("flags", *hud_panel_meter_flags),
    SInt8("minimum_meter_value"),
    SInt16("sequence_index"),
    SInt8("alpha_multiplier"),
    SInt8("alpha_bias"),
    SInt16("value_scale"),
    Float("opacity"),
    Float("translucency"),
    #QStruct("disabled_color", INCLUDE=argb_byte),
    UInt32("disabled_color", INCLUDE=argb_uint32),
    Pad(16),
    #QStruct("medium_health_left_color", INCLUDE=xrgb_byte),
    UInt32("medium_health_left_color", INCLUDE=xrgb_uint32),
    Float("max_color_health_fraction_cutoff"),
    Float("min_color_health_fraction_cutoff"),
    SIZE=136
    )

motion_sensor_center = Struct("motion_sensor_center",
    QStruct("anchor_offset",
        SInt16("x"), SInt16("y"), ORIENT='h',
        ),
    Float("width_scale"),
    Float("height_scale"),
    Bool16("scaling_flags", *hud_scaling_flags),
    SIZE=16
    )

auxilary_overlay = Struct("auxilary_overlay",
    Struct("background", INCLUDE=hud_background),
    SEnum16("type",
        "team_icon"
        ),
    Bool16("flags",
        "use_team_color"
        ),
    SIZE=132
    )

auxilary_meter = Struct("auxilary_meter",
    Pad(18),
    SEnum16("type", "integrated_light", VISIBLE=False),
    Struct("background", INCLUDE=hud_background),

    QStruct("anchor_offset",
        SInt16("x"), SInt16("y"), ORIENT='h',
        ),
    Float("width_scale"),
    Float("height_scale"),
    Bool16("scaling_flags", *hud_scaling_flags),

    Pad(22),
    dependency("meter_bitmap", "bitm"),
    #QStruct("color_at_meter_minimum", INCLUDE=xrgb_byte),
    #QStruct("color_at_meter_maximum", INCLUDE=xrgb_byte),
    #QStruct("flash_color", INCLUDE=xrgb_byte),
    #QStruct("empty_color", INCLUDE=argb_byte),
    UInt32("color_at_meter_minimum", INCLUDE=xrgb_uint32),
    UInt32("color_at_meter_maximum", INCLUDE=xrgb_uint32),
    UInt32("flash_color", INCLUDE=xrgb_uint32),
    UInt32("empty_color", INCLUDE=argb_uint32),
    Bool8("flags", *hud_panel_meter_flags),
    SInt8("minimum_meter_value"),
    SInt16("sequence_index"),
    SInt8("alpha_multiplier"),
    SInt8("alpha_bias"),
    SInt16("value_scale"),
    Float("opacity"),
    Float("translucency"),
    #QStruct("disabled_color", INCLUDE=argb_byte),
    UInt32("disabled_color", INCLUDE=argb_uint32),

    Pad(16),
    Float("minimum_fraction_cutoff"),
    Bool32("overlay_flags",
        "show_only_when_active",
        "flash_once_if_activated_while_disabled",
        ),

    SIZE=324,
    COMMENT="\nThis auxilary meter is meant for the flashlight.\n"
    )

unhi_body = Struct("tagdata",
    SEnum16("anchor", *hud_anchors),

    Pad(34),
    Struct("unit_hud_background", INCLUDE=hud_background),
    Struct("shield_panel_background", INCLUDE=hud_background),
    shield_panel_meter,
    Struct("health_panel_background", INCLUDE=hud_background),
    health_panel_meter,
    Struct("motion_sensor_background", INCLUDE=hud_background),
    Struct("motion_sensor_foreground", INCLUDE=hud_background),

    Pad(32),
    motion_sensor_center,

    Pad(20),
    SEnum16("auxilary_overlay_anchor", *hud_anchors),
    Pad(34),
    reflexive("auxilary_overlays", auxilary_overlay, 16),

    Pad(16),
    reflexive("warning_sounds", warning_sound, 12,
        DYN_NAME_PATH='.sound.filepath'),
    reflexive("auxilary_meters", auxilary_meter, 16),

    SIZE=1388
    )


def get():
    return unhi_def

unhi_def = TagDef("unhi",
    blam_header("unhi"),
    unhi_body,

    ext=".unit_hud_interface", endian=">", tag_cls=HekTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef


meter_body = Struct("tagdata",
    Pad(4),
    dependency("stencil_bitmap", "bitm"),
    dependency("source_bitmap",  "bitm"),

    SInt16("stencil_sequence_index"),
    SInt16("source_sequence_index"),
    Pad(20),
    SEnum16("interpolate_colors",
        "linearly",
        "faster_near_empty",
        "faster_near_full",
        "through_random_noise"
        ),
    SEnum16("anchor_colors" ,
        "at_both_ends",
        "at_empty",
        "at_full"
        ),
    Pad(8),
    QStruct("empty_color", INCLUDE=argb_float),
    QStruct("full_color",  INCLUDE=argb_float),
    Pad(20),
    Float("unmask_distance", SIDETIP="meter units"),
    Float("mask_distance", SIDETIP="meter units"),
    Pad(12),
    FlUInt16("screen_x_pos", SIDETIP="pixels"),
    FlUInt16("screen_y_pos", SIDETIP="pixels"),
    FlUInt16("width", SIDETIP="pixels"),
    FlUInt16("height", SIDETIP="pixels"),

    rawdata_ref("meter_data", max_size=65536),
    SIZE=172, WIDGET=MeterImageFrame
    )

def get():
    return metr_def

metr_def = TagDef("metr",
    blam_header('metr'),
    meter_body,
    ext=".meter", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


mffn_location = Struct("location",
    h3_string_id("name"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=8
    )


mffn_unknown_2 = Struct("unknown_2",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    Float("unknown_3", VISIBLE=False),
    Float("unknown_4", VISIBLE=False),
    Float("unknown_5", VISIBLE=False),
    h3_rawdata_ref("unknown_6"),
    Float("unknown_7", VISIBLE=False),
    h3_rawdata_ref("unknown_8"),
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=8, VISIBLE=False),
    h3_dependency("effect"),
    ENDIAN=">", SIZE=112
    )


mffn_body = Struct("tagdata",
    h3_dependency("render_model"),
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    SInt32("unknown_1", VISIBLE=False),
    h3_reflexive("locations", mffn_location),
    h3_reflexive("unknown_2", mffn_unknown_2),
    ENDIAN=">", SIZE=56
    )


def get():
    return mffn_def

mffn_def = TagDef("mffn",
    h3_blam_header('mffn'),
    mffn_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["mffn"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DeadCanadian
# 	Made soft barriers visable for easy access
# revision: 3		author: Lord Zedd
# 	ctrl c ctrl v
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


sddt_design_mopp_code_data = Struct("data",
    UInt8("data_byte"),
    ENDIAN=">", SIZE=1
    )


sddt_design_mopp_code = Struct("design_mopp_code",
    SInt32("unknown_0", VISIBLE=False),
    SInt16("size_0"),
    SInt16("size_1"),
    SInt32("offset_0"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    QStruct("offset_1", INCLUDE=xyz_float),
    Float("offset_scale"),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    SInt32("data_size"),
    UInt32("data_capacity"),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    h3_reflexive("data", sddt_design_mopp_code_data),
    BytesRaw("unknown_3", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=64
    )


sddt_design_shapes_2_unknown_2 = Struct("unknown_2",
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=17),
    VISIBLE=False,
    ENDIAN=">", SIZE=68
    )


sddt_design_shapes_2 = Struct("design_shapes_2",
    h3_string_id("name"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    h3_reflexive("unknown_2", sddt_design_shapes_2_unknown_2),
    ENDIAN=">", SIZE=20
    )


sddt_water_mopp_code_data = Struct("data",
    UInt8("data_byte"),
    ENDIAN=">", SIZE=1
    )


sddt_water_mopp_code = Struct("water_mopp_code",
    SInt32("unknown_0", VISIBLE=False),
    SInt16("size_0"),
    SInt16("size_1"),
    SInt32("offset_0"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    QStruct("offset_1", INCLUDE=xyz_float),
    Float("offset_scale"),
    BytesRaw("unknown_2", SIZE=4, VISIBLE=False),
    SInt32("data_size"),
    UInt32("data_capacity"),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    h3_reflexive("data", sddt_water_mopp_code_data),
    BytesRaw("unknown_3", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=64
    )


sddt_water_name = Struct("water_name",
    h3_string_id("name"),
    ENDIAN=">", SIZE=4
    )


sddt_underwater_definition_unknown_1 = Struct("unknown_1",
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=4),
    ENDIAN=">", SIZE=16
    )


sddt_underwater_definition_triangle = Struct("triangle",
    BytesRaw("_1_x", SIZE=36, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=36
    )


sddt_underwater_definition = Struct("underwater_definition",
    SInt16("water_name_index"),
    SInt16("unknown_0", VISIBLE=False),
    QStruct("flow_force", INCLUDE=xyz_float),
    Float("flow_force_z_2"),
    h3_reflexive("unknown_1", sddt_underwater_definition_unknown_1),
    h3_reflexive("triangles", sddt_underwater_definition_triangle),
    ENDIAN=">", SIZE=44
    )


sddt_body = Struct("tagdata",
    SInt32("unknown", VISIBLE=False),
    h3_reflexive("design_mopp_codes", sddt_design_mopp_code),
    h3_reflexive("design_shapes_2", sddt_design_shapes_2),
    h3_reflexive("water_mopp_codes", sddt_water_mopp_code),
    h3_reflexive("water_names", sddt_water_name),
    h3_reflexive("underwater_definitions", sddt_underwater_definition),
    ENDIAN=">", SIZE=64
    )


def get():
    return sddt_def

sddt_def = TagDef("sddt",
    h3_blam_header('sddt'),
    sddt_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["sddt"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from math import ceil, log, sqrt
from sys import float_info
from reclaimer.util import matrices


def point_distance_to_plane(point, plane, use_double_rounding=False,
                            round_adjust=0):
    mantissa_len = (53 if use_double_rounding else 23)
    # return True if point is on forward side of plane, otherwise False
    # take into account rounding errors for 32bit floats
    delta_max = 2**(
        int(log(abs(plane[3]) + float_info.epsilon, 2)) -
        mantissa_len) + abs(round_adjust)
    delta = matrices.dot_product(plane[:3], point) - plane[3]
    if abs(delta) < delta_max:
        # point lies on plane
        return 0.0
    return delta


def point_distance_to_line(point, line, use_double_rounding=False,
                           round_adjust=0):
    mantissa_len = (53 if use_double_rounding else 23)
    line_point = matrices.Point(line[0])
    line_dir   = matrices.Ray(line[1]).normalized
    dist = matrices.Ray.cross(line_dir, line_point - point).mag

    # return True if point is on forward side of plane, otherwise False
    # take into account rounding errors for 32bit floats
    delta_max = 2**(
        int(ceil(log(abs(dist) + float_info.epsilon, 2))) -
        mantissa_len) + abs(round_adjust)
    if abs(dist) < delta_max:
        # point lies on plane
        return 0.0
    return dist


def is_point_on_plane(point, plane, use_double_rounding=False,
                      round_adjust=0):
    return point_distance_to_plane(
        point, plane, use_double_rounding, round_adjust) == 0


def is_point_on_line(point, line, use_double_rounding=False,
                     round_adjust=0):
    return point_distance_to_line(
        point, line, use_double_rounding, round_adjust) == 0


def is_point_on_front_side_of_plane(point, plane, on_plane_ok=False,
                                    use_double_rounding=False,
                                    round_adjust=0):
    '''
    Returns if point iterable(x,y,z) is on the front side of
    plane iterable(x,y,z,delta).
    Either with single or double rounding.
    '''
    rounded_dist = point_distance_to_plane(
        point, plane, use_double_rounding, round_adjust)
    if on_plane_ok:
        return rounded_dist >= 0
    return rounded_dist > 0


def is_point_on_back_side_of_plane(point, plane, on_plane_ok=False,
                                   use_double_rounding=False,
                                   round_adjust=0):
    rounded_dist = point_distance_to_plane(
        point, plane, use_double_rounding, round_adjust) < 0
    if on_plane_ok:
        return rounded_dist <= 0
    return rounded_dist < 0


def is_point_on_front_side_of_planes(point, planes, on_plane_ok=False,
                                     use_double_rounding=False,
                                     round_adjust=0):
    for plane in planes:
        if not is_point_on_front_side_of_plane(
            point, plane, on_plane_ok, use_double_rounding, round_adjust):
            return False
    return True


def is_point_on_back_side_of_planes(point, planes, on_plane_ok=False,
                                    use_double_rounding=False,
                                    round_adjust=0):
    for plane in planes:
        if not is_point_on_back_side_of_plane(
            point, plane, on_plane_ok, use_double_rounding, round_adjust):
            return False
    return True


def find_intersect_line_of_planes(plane_0, plane_1, use_double_rounding=False,
                                  round_adjust=0, normalize=False):
    # returns an arbitrary point where the provided planes intersect.
    # if they do not intersect, None will be returned instead.
    plane_0 = Plane(plane_0)
    plane_1 = Plane(plane_1)
    plane_0.normalize()
    plane_1.normalize()
    plane_0_dir = matrices.Ray(plane_0[:3])
    plane_1_dir = matrices.Ray(plane_1[:3])
    line_dir = matrices.Ray.cross(plane_0_dir, plane_1_dir)
    determinant = line_dir.mag_sq
    if determinant <= float_info.epsilon:
        # planes are parallel. ignore
        return

    line_point = (
        (matrices.Ray.cross(line_dir, plane_0_dir) * plane_1[3]) +
        (matrices.Ray.cross(plane_1_dir, line_dir) * plane_0[3])) / determinant

    max_comp = 2
    if abs(line_dir[0]) >= abs(line_dir[1]):
        if abs(line_dir[0]) >= abs(line_dir[2]):
            max_comp = 0
    elif abs(line_dir[1]) >= abs(line_dir[2]):
        max_comp = 1

    line_point -= line_dir * (line_point[max_comp] / line_dir[max_comp])
    if normalize:
        line_dir.normalize()

    return line_point, line_dir


def find_intersect_point_of_lines(line_0, line_1, use_double_rounding=False,
                                  round_adjust=0, ignore_check=False):
    # returns an arbitrary point where the provided lines intersect.
    # if they do not intersect, None will be returned instead.
    v0, d0 = matrices.Point(line_0[0]), matrices.Ray(line_0[1])
    v1, d1 = matrices.Point(line_1[0]), matrices.Ray(line_1[1])
    assert len(v0) == len(v1)
    assert len(d0) == len(d1)
    if d0.is_zero or d1.is_zero:
       # one or both vectors are the zero-vector
        return None

    try:
        reduced, result = matrices.Matrix(
            [(a, -b) for a, b in zip(d0, d1)]).row_reduce(
            matrices.Matrix([v0[i] - v1[i] for i in range(len(v0))]))
    except matrices.CannotRowReduce:
        return None

    intersect_a = v0 - d0 * result[0][0]
    intersect_b = v1 - d1 * result[1][0]
    if ignore_check or matrices.are_vectors_equal(
            intersect_a, intersect_b, use_double_rounding, round_adjust):
        # point lies on both lines
        return intersect_a + (intersect_b - intersect_a) / 2

    return None


def find_intersect_point_of_planes(plane_0, plane_1, plane_2, use_double_rounding=False,
                                   round_adjust=0, ignore_check=False):
    try:
        reduced, result = matrices.Matrix(
            (plane_0[:-1], plane_1[:-1], plane_2[:-1])).row_reduce(
                matrices.Matrix([plane_0[3], plane_1[3], plane_2[3]]))
    except matrices.CannotRowReduce:
        return None

    return matrices.Point((result[0][0], result[1][0], result[2][0]))


def plane_from_verts(v0, v1, v2):
    p_dir = matrices.Ray.cross(matrices.line_from_verts(v1, v0),
                               matrices.line_from_verts(v1, v2))
    p_dir.normalize()
    if p_dir.is_zero:
        return None

    return (p_dir[0], p_dir[1], p_dir[2],
            v1[0]*p_dir[0] + v1[1]*p_dir[1] + v1[2]*p_dir[2])


def are_planes_equal(plane_0, plane_1, use_double_rounding=False,
                     round_adjust=0):
    p0 = Plane(plane_0)
    p1 = Plane(plane_1)
    p0.normalize()
    p1.normalize()
    return matrices.are_vectors_equal(p0, p1, use_double_rounding, round_adjust)


class Plane(matrices.FixedLengthList, matrices.Ray):
    __slots__ = ()
    def __init__(self, initializer=(0, 0, 1, 0)):
        assert len(initializer) == 4
        list.__init__(self, initializer)
    def __eq__(self, other):
        return are_planes_equal(self, other)
    @property
    def is_zero(self): return not(self[0] or self[1] or self[2])
    @property
    def magnitude_sq(self): return self[0]**2 + self[1]**2 + self[2]**2
    @property
    def magnitude(self): return sqrt(self[0]**2 + self[1]**2 + self[2]**2)
    mag_sq = magnitude_sq
    mag = magnitude

    @property
    def i(self): return self[0]
    @i.setter
    def i(self, new_val): self[0] = float(new_val)
    @property
    def j(self): return self[1]
    @j.setter
    def j(self, new_val): self[1] = float(new_val)
    @property
    def k(self): return self[2]
    @k.setter
    def k(self, new_val): self[2] = float(new_val)
    @property
    def d(self): return self[3]
    @d.setter
    def d(self, new_val): self[3] = float(new_val)

    def normalize(self):
        div = self.mag
        if div:
            self[0] /= div
            self[1] /= div
            self[2] /= div
            self[3] *= div

    def is_point_on_plane(self, point, use_double_rounding=False,
                          round_adjust=0):
        return is_point_on_plane(
            point, self, use_double_rounding, round_adjust)

    def is_point_on_plane_front(self, point, on_plane_ok=False,
                                use_double_rounding=False, round_adjust=0):
        return is_point_on_front_side_of_plane(
            point, self, on_plane_ok, use_double_rounding, round_adjust)

    def is_point_on_plane_front(self, point, on_plane_ok=False,
                                use_double_rounding=False, round_adjust=0):
        return is_point_on_back_side_of_plane(
            point, self, on_plane_ok, use_double_rounding, round_adjust)

    def find_intersect_line_with_plane(self, other, use_double_rounding=False,
                                       round_adjust=0):
        return find_intersect_line_of_planes(
            self, other, use_double_rounding, round_adjust)


class EdgeLoopNode(list):
    __slots__ = ("vert_index", )

    def __init__(self, vert_index=-1, *children, edges=(), seen=None):
        self.vert_index = int(vert_index)
        if seen is None:
            seen = set()

        if self.vert_index in seen:
            return

        seen.add(self.vert_index)

        if not isinstance(edges, dict):
            edges_by_verts = {}
            for edge in set(edges):
                edges_by_verts.setdefault(edge[0], []).append(edge)
                edges_by_verts.setdefault(edge[1], []).append(edge)

            for vert_index in edges_by_verts:
                self.append(EdgeLoopNode(
                    vert_index, edges=edges_by_verts))
        else:
            for edge in edges.get(self.vert_index, ()):
                if edge[0] == self.vert_index:
                    self.append(EdgeLoopNode(edge[1], edges=edges, seen=set(seen)))

    def __str__(self, **kw):
        depth = kw.pop("depth", 0)
        kw["depth"] = depth + 1
        indent_str = " " * (2 * depth)

        string = "%sEdgeLoopNode(%s" % (indent_str, self.vert_index)
        if not self.is_leaf:
            string += ",\n"
            for node in self:
                string += node.__str__(**kw) + ",\n"
            string += indent_str
        return string + ")"

    __repr__ = __str__

    def walk_edge_loops(self, curr_loop=(), seen=None, horizontal=False):
        if seen is None:
            seen = set()

        if (not horizontal or self.is_leaf) and self.vert_index != -1:
            curr_loop += (self.vert_index, )

        if self.vert_index in seen:
            yield curr_loop
            return

        seen.add(self.vert_index)
        if self.is_leaf:
            yield curr_loop
            return

        # walk down the length of the loop
        for node in self:
            yield from node.walk_edge_loops(curr_loop, seen, horizontal)

        if self.vert_index == -1 or not horizontal:
            return

        # walk horizontally across the loops
        for i in range(len(self)):
            for j in range(len(self)):
                if i == j: continue
                seen_h = {self.vert_index}
                for loop in self[i].walk_edge_loops((self.vert_index, ), seen_h):
                    yield from self[j].walk_edge_loops(loop[::-1], set(seen_h))

    @property
    def is_leaf(self): return not len(self)

    @property
    def depth(self):
        if self.is_leaf:
            return 0
        depth = max(n.depth for n in self)
        if self.vert_index != -1:
            depth += 1
        return depth


def planes_to_verts_and_edge_loops(planes, center, plane_dir=True, max_plane_ct=32,
                                   use_double_rounding=False, round_adjust=0):
    assert len(center) == 3
    # make a set out of the planes to remove duplicates
    planes = list(set(tuple(Plane(p).normalized) for p in planes))
    indices_by_planes = {p: set() for p in planes}

    if len(planes) > max_plane_ct:
        raise ValueError(
            "Provided more planes(%s) than the max plane count(%s)" %
            (len(planes), max_plane_ct))

    is_point_inside_form = (is_point_on_front_side_of_planes if plane_dir
                            else is_point_on_back_side_of_planes)

    # find the intersect points of all planes within the polyhedron
    intersects = []
    intersect_weights = []
    for i in range(len(planes)):
        p0 = planes[i]
        p0_indices = indices_by_planes[p0]
        for j in range(len(planes)):
            if j == i: continue
            p1 = planes[j]
            p1_indices = indices_by_planes[p1]
            for k in range(len(planes)):
                if k == i or k == j: continue
                p2 = planes[k]
                p2_indices = indices_by_planes[p2]

                intersect = find_intersect_point_of_planes(p0, p1, p2)
                if intersect is None or not is_point_inside_form(
                        intersect, planes, True, round_adjust=round_adjust):
                    continue

                weight = abs(matrices.Ray.cross(p0[: -1], p1[: -1]) *
                             matrices.Ray(p2[: -1]))
                better_vert_index = None
                similar_verts = set()
                for w in range(len(intersects)):
                    if matrices.are_vectors_equal(
                            intersects[w], intersect,
                            round_adjust=round_adjust / weight):
                        if intersect_weights[w] > weight:
                            better_vert_index = w
                        else:
                            similar_verts.add(w)

                if better_vert_index is None:
                    # adding a new vert
                    vert_index = len(intersects)
                    intersects.append(intersect)
                    intersect_weights.append(weight)
                else:
                    # using a better intersect than this one
                    vert_index = better_vert_index
                    intersect = intersects[better_vert_index]
                    weight = intersect_weights[better_vert_index]

                # overwrite existing verts with a more accurate one
                for v_i in similar_verts:
                    intersects[v_i] = intersect
                    intersect_weights[v_i] = weight

                p0_indices.add(vert_index)
                p1_indices.add(vert_index)
                p2_indices.add(vert_index)

    vert_map = {}
    vert_rebase_map = [None] * len(intersects)
    pruned_intersects = []
    pruned_intersect_weights = []
    for i in range(len(intersects)):
        vert = tuple(intersects[i])
        if vert not in vert_map:
            vert_map[vert] = len(vert_map)
            pruned_intersects.append(intersects[i])
            pruned_intersect_weights.append(intersect_weights[i])
        vert_rebase_map[i] = vert_map[vert]

    for indices in indices_by_planes.values():
        new_indices = set()
        for i in indices:
            new_indices.add(vert_rebase_map[i])
        indices.clear()
        indices.update(new_indices)

    intersects = pruned_intersects
    intersect_weights = pruned_intersect_weights
    #for i in range(len(intersects)):
    #    print(intersect_weights[i], intersects[i])

    # Get intersection lines by crossing each plane with each other plane.
    lines_by_planes = {plane: set() for plane in planes}
    for p0_i in range(len(planes) - 1):
        for p1_i in range(p0_i + 1, len(planes)):
            p0 = planes[p0_i]
            p1 = planes[p1_i]

            line = find_intersect_line_of_planes(p0, p1)
            if line is None:
                continue

            lines_by_planes[p0].add((tuple(line[0]), tuple(line[1])))
            lines_by_planes[p1].add((tuple(line[0]), tuple(-line[1])))


    edges_by_planes = {plane: set() for plane in planes}
    used_vert_indices = set()
    for plane, lines in lines_by_planes.items():
        used_vert_counts = {i: [] for i in indices_by_planes[plane]}

        # count how many lines each vert lies on
        for line in lines:
            line_weight = matrices.Ray(line[1]).mag
            vert_indices_on_line = set()
            for v_i in used_vert_counts:
                if is_point_on_line(intersects[v_i], line,
                                    round_adjust=(round_adjust/(
                                        intersect_weights[v_i] * line_weight))):
                    vert_indices_on_line.add(v_i)

            # filter out verts only intersecting one line
            if len(vert_indices_on_line) >= 2:
                for v_i in vert_indices_on_line:
                    used_vert_counts[v_i].append(line)

        verts_on_lines = {line: [] for line in lines}
        for v_i, used_lines in used_vert_counts.items():
            if len(used_lines) < 2: continue
            for line in used_lines:
                verts_on_lines[line].append(v_i)

        for line in lines:
            if len(verts_on_lines[line]) < 2:
                continue

            line_point, line_dir = line
            sorted_verts_on_line = {}
            axis = 0
            # figure out which axis has the most precision
            for i in range(1, len(line_dir)):
                if abs(line_dir[i]) > abs(line_dir[axis]):
                    axis = i

            # sort the vertices by how far along the line they are
            for v_i in verts_on_lines[line]:
                axis_val = intersects[v_i][axis]
                dist = (axis_val - line_point[axis]) / line_dir[axis]
                sorted_verts_on_line[dist] = v_i

            verts_on_line = [sorted_verts_on_line[dist]
                             for dist in sorted(sorted_verts_on_line)]

            # if the edge chain will go opposite the line, reverse it
            if matrices.Ray(intersects[verts_on_line[0]] -
                            intersects[verts_on_line[-1]]) * matrices.Ray(line[1]) < 0:
                verts_on_line = verts_on_line[::-1]

            # generate edges that ascendingly follow this line
            for i in range(len(verts_on_line) - 1):
                v0_i = verts_on_line[i]
                v1_i = verts_on_line[i + 1]
                # planes intersect and are not parallel
                edges_by_planes[plane].add((v0_i, v1_i))
                used_vert_indices.update((v0_i, v1_i))


    # prune the unused verts
    verts = [None] * len(used_vert_indices)
    vert_map = {}
    i = 0
    for vert_index in used_vert_indices:
        vert_map[vert_index] = i
        verts[i] = intersects[vert_index]
        i += 1

    for plane, edges in edges_by_planes.items():
        new_edges = set()
        for edge in edges:
            new_edges.add((vert_map[edge[0]], vert_map[edge[1]]))
        edges.clear()
        edges.update(new_edges)

    edge_loops = []
    # loop over each edge and put together an edge loop list
    # by visiting the edges shared by each vert index
    for plane, edges in edges_by_planes.items():
        if not edges:
            continue

        try:
            edge_loop = edges_to_edge_loop(edges, plane, verts)
        except Exception:
            from traceback import format_exc
            print(format_exc())
            continue

        if not edge_loop:
            continue

        edge_loops.append(edge_loop)

    #print("PLANES:", )
    #i = 0
    #for plane in edges_by_planes:
    #    print(i, plane)
    #    for e in edges_by_planes[plane]:
    #        print(e)
    #    print()
    #    i += 1

    #i = 0
    #print("VERTS:", len(verts))
    #for vert in verts:
    #    print(i, vert)
    #    i += 1

    #print("TRIS:", sum(len(edge_loop) - 2 for edge_loop in edge_loops))
    #for edge_loop in edge_loops:
    #    print(edge_loop)

    return verts, edge_loops


def get_is_valid_edge_loop(edge_loop, plane, verts):
    vert_ct = len(edge_loop)
    if vert_ct < 3:
        return False

    # if the edge loop would construct triangles facing the
    # wrong direction, we need to reverse the edge loop
    for i in range(vert_ct):
        tri_plane = matrices.Ray(matrices.vertex_cross_product(
            verts[edge_loop[i]],
            verts[edge_loop[(i + 1) % vert_ct]],
            verts[edge_loop[(i + 2) % vert_ct]]))

        if matrices.dot_product(tri_plane, plane[: 3]) < 0:
            # facing opposite directions
            return False

    return True


def edges_to_edge_loop(edges, plane, verts):
    edge_loop_nodes = EdgeLoopNode(-1, edges=edges)
    #print("DEPTH: ", edge_loop_nodes.depth)
    #print(edges)
    #input(edge_loop_nodes)
    edge_loop = ()
    for curr_loop in edge_loop_nodes.walk_edge_loops():
        if len(curr_loop) < 3 or curr_loop[0] != curr_loop[-1]:
            continue

        curr_loop = curr_loop[: -1]
        if len(curr_loop) < len(edge_loop):
            continue
        elif get_is_valid_edge_loop(curr_loop, plane, verts):
            edge_loop = curr_loop

    for curr_loop in edge_loop_nodes.walk_edge_loops(horizontal=True):
        if len(curr_loop) < 3 or curr_loop[0] != curr_loop[-1]:
            continue

        curr_loop = curr_loop[: -1]
        if len(curr_loop) < len(edge_loop):
            continue
        elif get_is_valid_edge_loop(curr_loop, plane, verts):
            edge_loop = curr_loop

    #print("CHOSEN: ", edge_loop)
    #print()
    if len(edge_loop) < 3:
        return None

    return list(edge_loop)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.mod2 import Mod2Tag
from supyr_struct.defs.tag_def import TagDef

def get():
    return mod2_def

local_marker = Struct('local_marker',
    ascii_str32("name"),
    dyn_senum16('node_index',
        DYN_NAME_PATH="tagdata.nodes.nodes_array[DYN_I].name"),
    Pad(2),

    QStruct('rotation', INCLUDE=ijkw_float),
    QStruct('translation', INCLUDE=xyz_float),
    SIZE=80
    )

fast_uncompressed_vertex = QStruct('uncompressed_vertex',
    Float('position_x'), Float('position_y'), Float('position_z'),
    Float('normal_i'),   Float('normal_j'),   Float('normal_k'),
    Float('binormal_i'), Float('binormal_j'), Float('binormal_k'),
    Float('tangent_i'),  Float('tangent_j'),  Float('tangent_k'),

    Float('u'), Float('v'),

    SInt16('node_0_index'), SInt16('node_1_index'),
    Float('node_0_weight'), Float('node_1_weight'),
    SIZE=68
    )

fast_compressed_vertex = QStruct('compressed_vertex',
    Float('position_x'), Float('position_y'), Float('position_z'),
    UInt32('normal'),
    UInt32('binormal'),
    UInt32('tangent'),

    SInt16('u', UNIT_SCALE=1/32767, MIN=-32767, WIDGET_WIDTH=10),
    SInt16('v', UNIT_SCALE=1/32767, MIN=-32767, WIDGET_WIDTH=10),

    SInt8('node_0_index', UNIT_SCALE=1/3, MIN=0, WIDGET_WIDTH=10),
    SInt8('node_1_index', UNIT_SCALE=1/3, MIN=0, WIDGET_WIDTH=10),
    SInt16('node_0_weight', UNIT_SCALE=1/32767, MIN=0, WIDGET_WIDTH=10),
    SIZE=32
    )

uncompressed_vertex = Struct('uncompressed_vertex',
    QStruct("position", INCLUDE=xyz_float),
    QStruct("normal", INCLUDE=ijk_float),
    QStruct("binormal", INCLUDE=ijk_float),
    QStruct("tangent", INCLUDE=ijk_float),

    Float('u'),
    Float('v'),

    SInt16('node_0_index',
        TOOLTIP="If local nodes are used, this is a local index"),
    SInt16('node_1_index',
        TOOLTIP="If local nodes are used, this is a local index"),
    Float('node_0_weight'),
    Float('node_1_weight'),
    SIZE=68
    )

compressed_vertex = Struct('compressed_vertex',
    QStruct("position", INCLUDE=xyz_float),

    BBitStruct('normal',   INCLUDE=compressed_normal_32, SIZE=4),
    BBitStruct('binormal', INCLUDE=compressed_normal_32, SIZE=4),
    BBitStruct('tangent',  INCLUDE=compressed_normal_32, SIZE=4),

    SInt16('u', UNIT_SCALE=1/32767, MIN=-32767, WIDGET_WIDTH=10),
    SInt16('v', UNIT_SCALE=1/32767, MIN=-32767, WIDGET_WIDTH=10),

    SInt8('node_0_index', UNIT_SCALE=1/3, MIN=0, WIDGET_WIDTH=10),
    SInt8('node_1_index', UNIT_SCALE=1/3, MIN=0, WIDGET_WIDTH=10),
    SInt16('node_0_weight', UNIT_SCALE=1/32767, MIN=0, WIDGET_WIDTH=10),
    SIZE=32
    )

triangle = QStruct('triangle',
    SInt16('v0_index'), SInt16('v1_index'), SInt16('v2_index'),
    SIZE=6, ORIENT='h'
    )

uncompressed_vertex_union = Union('uncompressed_vertex',
    CASES={'uncompressed_vertex': uncompressed_vertex},
    )

compressed_vertex_union = Union('compressed_vertex',
    CASES={'compressed_vertex': compressed_vertex},
    )

triangle_union = Union('triangle',
    CASES={'triangle': triangle},
    )


marker_instance = Struct('marker_instance',
    dyn_senum8('region_index',
        DYN_NAME_PATH="tagdata.regions.regions_array[DYN_I].name"),
    SInt8('permutation_index', MIN=-1),
    dyn_senum8('node_index',
        DYN_NAME_PATH="tagdata.nodes.nodes_array[DYN_I].name"),
    Pad(1),

    QStruct('translation', INCLUDE=xyz_float),
    QStruct('rotation', INCLUDE=ijkw_float),
    SIZE=32
    )

permutation = Struct('permutation',
    ascii_str32("name"),
    Bool32('flags',
        'cannot_be_chosen_randomly'
        ),
    # permutations ending with -XXX where XXX is a number will belong to
    # the permutation set XXX. Trailing non-numeric characters are ignored.
    # Example:
    #     head_marcus_cap-101asdf
    #     Will have its permutation_set be set to 101
    # For all parts of a permutation to be properly chosen across
    # all regions, they must share a permutation set number.
    # Anything not ending in -XXX will have this set to 0.
    FlUInt16("permutation_set", VISIBLE=False),  # meta only field
    Pad(26),

    dyn_senum16('superlow_geometry_block',
        DYN_NAME_PATH="tagdata.geometries.geometries_array[DYN_I].NAME"),
    dyn_senum16('low_geometry_block',
        DYN_NAME_PATH="tagdata.geometries.geometries_array[DYN_I].NAME"),
    dyn_senum16('medium_geometry_block',
        DYN_NAME_PATH="tagdata.geometries.geometries_array[DYN_I].NAME"),
    dyn_senum16('high_geometry_block',
        DYN_NAME_PATH="tagdata.geometries.geometries_array[DYN_I].NAME"),
    dyn_senum16('superhigh_geometry_block',
        DYN_NAME_PATH="tagdata.geometries.geometries_array[DYN_I].NAME"),
    Pad(2),

    reflexive("local_markers", local_marker, 32, DYN_NAME_PATH=".name"),
    SIZE=88
    )

part = Struct('part',
    Bool32('flags',
        'stripped',
        'ZONER',
        ),
    dyn_senum16('shader_index',
        DYN_NAME_PATH="tagdata.shaders.shaders_array[DYN_I].shader.filepath"),
    SInt8('previous_part_index'),
    SInt8('next_part_index'),

    SInt16('centroid_primary_node'),
    SInt16('centroid_secondary_node'),
    Float('centroid_primary_weight'),
    Float('centroid_secondary_weight'),

    QStruct('centroid_translation', INCLUDE=xyz_float),

    #reflexive("uncompressed_vertices", uncompressed_vertex_union, 32767),
    #reflexive("compressed_vertices", compressed_vertex_union, 32767),
    #reflexive("triangles", triangle_union, 32767),
    reflexive("uncompressed_vertices", fast_uncompressed_vertex, 32767),
    reflexive("compressed_vertices", fast_compressed_vertex, 32767),
    reflexive("triangles", triangle, 32767),
    #Pad(36),
    Struct("model_meta_info",
        UEnum16("index_type",  # name is a guess.  always 1?
            ("uncompressed", 1),
            ),
        Pad(2),
        UInt32("index_count"),
        # THESE TWO VALUES ARE DIFFERENT THAN ON XBOX IT SEEMS
        UInt32("indices_magic_offset"),
        UInt32("indices_offset"),

        UEnum16("vertex_type",  # name is a guess
            ("uncompressed", 4),
            ("compressed",   5),
            ),
        Pad(2),
        UInt32("vertex_count"),
        Pad(4),  # always 0?
        # THESE TWO VALUES ARE DIFFERENT THAN ON XBOX IT SEEMS
        UInt32("vertices_magic_offset"),
        UInt32("vertices_offset"),
        VISIBLE=False, SIZE=36
        ),

    Pad(3),
    SInt8('local_node_count', MIN=0, MAX=22),
    #UInt8Array('local_nodes', SIZE=22),
    Array("local_nodes", SUB_STRUCT=UInt8("local_node_index"), SIZE=22),

    # this COULD be 2 more potential local nodes, but I've seen tool
    # split models when they reach 22 nodes, so im assuming 22 is the max
    Pad(2),
    SIZE=132
    )

fast_part = dict(part)
fast_part[9]  = raw_reflexive("uncompressed_vertices", fast_uncompressed_vertex, 65535)
fast_part[10] = raw_reflexive("compressed_vertices", fast_compressed_vertex, 65535)
fast_part[11] = raw_reflexive("triangles", triangle, 65535)

marker = Struct('marker',
    ascii_str32("name"),
    UInt16('magic_identifier'),
    Pad(18),

    reflexive("marker_instances", marker_instance, 32),
    SIZE=64
    )

node = Struct('node',
    ascii_str32("name"),
    dyn_senum16('next_sibling_node', DYN_NAME_PATH="..[DYN_I].name"),
    dyn_senum16('first_child_node', DYN_NAME_PATH="..[DYN_I].name"),
    dyn_senum16('parent_node', DYN_NAME_PATH="..[DYN_I].name"),
    Pad(2),

    QStruct('translation', INCLUDE=xyz_float),
    QStruct('rotation', INCLUDE=ijkw_float),
    Float('distance_from_parent'),
    Pad(32),

    # xbox specific values
    LFloat('scale', ENDIAN='<', DEFAULT=1.0, VISIBLE=False),
    QStruct("rot_jj_kk", GUI_NAME="[1-2j^2-2k^2]   2[ij+kw]   2[ik-jw]",
        INCLUDE=ijk_float, ENDIAN='<', VISIBLE=False),
    QStruct("rot_kk_ii", GUI_NAME="2[ij-kw]   [1-2k^2-2i^2]   2[jk+iw]",
        INCLUDE=ijk_float, ENDIAN='<', VISIBLE=False),
    QStruct("rot_ii_jj", GUI_NAME="2[ik+jw]   2[jk-iw]   [1-2i^2-2j^2]",
        INCLUDE=ijk_float, ENDIAN='<', VISIBLE=False),
    QStruct('translation_to_root',
        INCLUDE=xyz_float, ENDIAN='<', VISIBLE=False),
    SIZE=156,
    )

region = Struct('region',
    ascii_str32("name"),
    Pad(32),
    reflexive("permutations", permutation, 32, DYN_NAME_PATH=".name"),
    SIZE=76
    )

geometry = Struct('geometry',
    Pad(36),
    reflexive("parts", part, 32),
    SIZE=48
    )

fast_geometry = Struct('geometry',
    Pad(36),
    reflexive("parts", fast_part, 32),
    SIZE=48
    )

shader = Struct('shader',
    dependency("shader", valid_shaders),
    SInt16('permutation_index'),
    SIZE=32,
    )


mod2_body = Struct('tagdata',
    Bool32('flags',
        'blend_shared_normals',
        'parts_have_local_nodes',
        'ignore_skinning'
        ),
    SInt32('node_list_checksum'),

    Float('superhigh_lod_cutoff', SIDETIP="pixels"),
    Float('high_lod_cutoff', SIDETIP="pixels"),
    Float('medium_lod_cutoff', SIDETIP="pixels"),
    Float('low_lod_cutoff', SIDETIP="pixels"),
    Float('superlow_lod_cutoff', SIDETIP="pixels"),

    SInt16('superlow_lod_nodes', SIDETIP="nodes"),
    SInt16('low_lod_nodes', SIDETIP="nodes"),
    SInt16('medium_lod_nodes', SIDETIP="nodes"),
    SInt16('high_lod_nodes', SIDETIP="nodes"),
    SInt16('superhigh_lod_nodes', SIDETIP="nodes"),

    Pad(10),

    Float('base_map_u_scale'),
    Float('base_map_v_scale'),

    Pad(116),

    reflexive("markers", marker, 256, DYN_NAME_PATH=".name"),
    reflexive("nodes", node, 64, DYN_NAME_PATH=".name"),
    reflexive("regions", region, 32, DYN_NAME_PATH=".name"),
    reflexive("geometries", geometry, 256),
    reflexive("shaders", shader, 256, DYN_NAME_PATH=".shader.filepath"),

    SIZE=232
    )

fast_mod2_body = dict(mod2_body)
fast_mod2_body[19] = reflexive("geometries", fast_geometry, 256)

mod2_def = TagDef("mod2",
    blam_header('mod2', 5),
    mod2_body,

    ext=".gbxmodel", endian=">", tag_cls=Mod2Tag
    )

fast_mod2_def = TagDef("mod2",
    blam_header('mod2', 5),
    fast_mod2_body,

    ext=".gbxmodel", endian=">", tag_cls=Mod2Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

class VorbisBitrateInfo:
    '''
    Intermediary class for storing bitrate info to pass to vorbis compression
    functions.

    Combinations of nominal, lower, upper values carry the following
    implications:
      all three set to the same value:
        implies a fixed rate bitstream
      only nominal set:
        implies a VBR stream that nominals the nominal bitrate. No hard
        upper/lower limit
      upper and or lower set:
        implies a VBR bitstream that obeys the bitrate limits. nominal
        may also be set to give a nominal rate.
      none set:
        the coder does not care to speculate.
    '''
    lower = -1
    upper = -1
    nominal = -1

    # compression base quality [-0.1, 1.0]
    quality = None

    use_quality = False

    def __init__(self, nominal=-1, lower=-1, upper=-1, quality=0.5):
        '''
        See class documentation for variable descriptions.
        '''
        if quality is not None:
            self.set_bitrate_quality(quality)
        else:
            self.set_bitrate_variable(nominal, upper, lower)

    def set_bitrate_fixed(self, bitrate):
        '''
        Sets a fixed bitrate to the requested number.
        '''
        self.upper = self.nominal = self.lower = bitrate
        self.use_quality = False

    def set_bitrate_variable(self, nominal, upper=-1, lower=-1):
        '''
        Sets a variable bitrate based on the nominal and
        optionally upper and lower numbers.
        Check class docstring for further detail.
        '''
        self.nominal = nominal
        self.upper = upper
        self.lower = lower
        self.use_quality = False

    def set_bitrate_quality(self, quality):
        '''
        Sets a bitrate using the quality float.
        '''
        self.quality = min(1.0, max(-0.1, float(quality)))
        self.use_quality = True



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Copypasta H2
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


snmx_body = Struct("tagdata",
    Float("left_stereo_gain_0"),
    Float("right_stereo_gain_0"),
    Float("left_stereo_gain_1"),
    Float("right_stereo_gain_1"),
    Float("left_stereo_gain_2"),
    Float("right_stereo_gain_2"),
    Float("front_speaker_gain_0"),
    Float("rear_speaker_gain_0"),
    Float("front_speaker_gain_1"),
    Float("rear_speaker_gain_1"),
    Float("mono_unspatialized_gain"),
    Float("stereo_to_3d_gain"),
    Float("rear_surround_to_front_stereo_gain"),
    Float("front_speaker_gain_2"),
    Float("center_speaker_gain_0"),
    Float("front_speaker_gain_3"),
    Float("center_speaker_gain_1"),
    Float("stereo_unspatialized_gain"),
    Float("solo_player_fade_out_delay"),
    Float("solo_player_fade_out_time"),
    Float("solo_player_fade_in_time"),
    Float("game_music_fade_out_time"),
    h3_dependency("unknown_0"),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=112
    )


def get():
    return snmx_def

snmx_def = TagDef("snmx",
    h3_blam_header('snmx'),
    snmx_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["snmx"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

def get(): return flag_def

attachment_point = Struct("attachment_point",
    SInt16("height_to_next_attachment", SIDETIP="vertices"),
    Pad(18),
    ascii_str32("marker_name"),
    )

flag_body = Struct("tagdata",
    Pad(4),
    SEnum16("trailing_edge_shape",
        "flat",
        "concave_triangular",
        "convex_triangular",
        "trapezoid_short_top",
        "trapezoid_short_bottom",
        ),

    SInt16("trailing_edge_shape_offset", SIDETIP="vertices"),
    SEnum16("attached_edge_shape",
        "flat",
        "concave_triangular",
        ),
    Pad(2),
    SInt16("width", SIDETIP="vertices"),
    SInt16("height", SIDETIP="vertices"),

    float_wu("cell_width"),
    float_wu("cell_height"),

    dependency("red_flag_shader", valid_shaders),
    dependency("physics", "pphy"),

    float_wu_sec("wind_noise"),
    Pad(8),
    dependency("blue_flag_shader", valid_shaders),
    reflexive("attachment_points", attachment_point, 4,
        DYN_NAME_PATH='.marker_name'),
    SIZE=96,
    )

flag_def = TagDef("flag",
    blam_header('flag'),
    flag_body,

    ext=".flag", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.pphy import PphyTag
from supyr_struct.defs.tag_def import TagDef

density_comment = """Densities(grams per milliliter)\n
air        0.0011
snow       0.128
cork       0.24
cedar      0.43
oak        0.866
ice        0.897
water      1.0
soil       1.1
cotton     1.491
dry earth  1.52
sand       1.7
granite    2.4
glass      2.5
iron       7.65
steel      7.77
lead       11.37
uranium    18.74
gold       19.3"""

pphy_body = Struct("tagdata",
    Bool32("flags",
        "flamethrower_particle_collision",
        "collides_with_structures",
        "collides_with_water_surface",
        "uses_simple_wind",
        "uses_damped_wind",
        "no_gravity"
        ),
    # these next three are courtesy of Sparky. I had
    # no idea these existed till I looked in Eschaton.
    # kavawuvi figured out how to calculate them(see PphyTag)
    FlFloat("scaled_density", VISIBLE=False),
    FlFloat("water_gravity_scale", VISIBLE=False),
    FlFloat("air_gravity_scale", VISIBLE=False),
    Pad(16),
    Float("density", SIDETIP="g/mL", COMMENT=density_comment),#g/mL
    Float("air_friction"),
    Float("water_friction"),
    Float("surface_friction"),
    Float("elasticity"),

    SIZE=64
    )

def get():
    return pphy_def

pphy_def = TagDef("pphy",
    blam_header('pphy'),
    pphy_body,

    ext=".point_physics", endian=">", tag_cls=PphyTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

# Typing these up took FOREVER
game_data_input_functions = (
    'NULL', 'player_settings_menu_update', 'unused',
    'playlist_settings_menu_update', 'gametype_select_menu_update',
    'multiplayer_type_menu_update', 'solo_level_select_update',
    'difficulty_menu_update', 'build_number',  # textbox only for build_number
    'server_list_update', 'network_pregame_status_update',
    'splitscreen_pregame_update', 'net_splitscreen_prejoin_players',
    'mp_profile_list_update', 'wide_3_player_profile_list_update',
    'player_profile_edit_select_menu_update',
    'player_profile_small_menu_update', 'game_settings_lists_text_update',
    'solo_game_objective_text', 'color_picker_update',
    'game_setings_lists_pic_update', 'main_menu_fake_animate',
    'mp_level_select_update',
    'get_active_player_profile_name', 'get_edit_player_profile_name',
    'get_edit_game_settings_name', 'get_active_player_profile_color',
    'mp_set_textbox_map_name', 'mp_set_textbox_game_rules',
    'mp_set_textbox_teams_no_teams', 'mp_set_textbox_score_limit',
    'mp_set_textbox_score_limit_type', 'mp_set_bitmap_for_map',
    'mp_set_bitmap_for_ruleset', 'mp_set_textbox_player_count',
    'mp_edit_profile_set_rule_text',
    'system_link_status_check', 'mp_game_directions',
    'teams_no_teams_bitmap_update', 'warn_if_diff_will_nuke_saved_game',
    'dim_if_no_net_cable', 'pause_game_set_textbox_inverted',
    'dim_unless_two_controllers', 'controls_update_menu',
    'video_menu_update', 'gamespy_screen_update', 'common_button_bar_update',
    'gamepad_update_menu', 'server_settings_update', 'audio_menu_update',
    'mp_profile_vehicles_update', 'solo_map_list_update', 'mp_map_list_update',
    'gametype_select_list_update', 'gametype_edit_list_update',
    'load_game_list_update', 'checking_for_updates',
    'direct_ip_connect_update', 'network_settings_update',
    )
event_types = (
    'A_button', 'B_button', 'X_button', 'Y_button',
    'black_button', 'white_button', 'left_trigger', 'right_trigger',
    'dpad_up', 'dpad_down', 'dpad_left', 'dpad_right',
    'start_button', 'back_button', 'left_thumb', 'right_thumb',
    'left_analog_stick_up', 'left_analog_stick_down',
    'left_analog_stick_left', 'left_analog_stick_right',
    'right_analog_stick_up', 'right_analog_stick_down',
    'right_analog_stick_left', 'right_analog_stick_right',
    'created', 'deleted', 'get_focus', 'lose_focus',
    'left_mouse', 'middle_mouse', 'right_mouse', 'double_click',
    'custom_activation', 'post_render'
    )
event_functions = (
    'NULL', 'list_goto_next_item', 'list_goto_previous_item',
    'unused1', 'unused2', 'initialize_sp_level_list_solo',
    'initialize_sp_level_list_coop', 'dispose_sp_level_list',
    'solo_level_set_map', 'set_difficulty', 'start_new_game',
    'pause_game_restart_at_checkpoint', 'pause_game_restart_level',
    'pause_game_rturn_to_main_menu', 'clear_multiplayer_player_joins',
    'join_controller_to_mp_game', 'initialize_net_game_server_list',
    'start_net_game_server', 'dispose_net_game_server_list',
    'shutdown_net_game_server', 'net_game_join_from_server_list',
    'split_screen_game_initialize', 'coop_game_initialize',
    'main_menu_initialize', 'mp_type_menu_initialize',
    'pick_play_stage_for_quick_start', 'mp_level_list_initialize',
    'mp_level_list_dispose', 'mp_level_select', 'mp_profiles_list_initialize',
    'mp_profiles_list_dispose', 'mp_profile_set_for_game', 'swap_player_team',
    'net_game_join_player', 'player_profile_list_initialize',
    'player_profile_list_dispose',
    'wide_3_player_profile_set_for_game', 'wide_1_player_profile_set_for_game',
    'mp_profile_begin_editing', 'mp_profile_end_editing',
    'mp_profile_set_game_engine', 'mp_profile_change_name',
    'mp_profile_set_ctf_rules', 'mp_profile_set_koth_rules',
    'mp_profile_set_slayer_rules', 'mp_profile_set_oddball_rules',
    'mp_profile_set_racing_rules', 'mp_profile_set_player_options',
    'mp_profile_set_item_options', 'mp_profile_set_indicator_options',
    'mp_profile_init_game_engine', 'mp_profile_init_name',
    'mp_profile_init_ctf_rules', 'mp_profile_init_koth_rules',
    'mp_profile_init_slayer_rules', 'mp_profile_init_oddball_rules',
    'mp_profile_init_racing_rules', 'mp_profile_init_player_options',
    'mp_profile_init_item_options', 'mp_profile_init_indicator_options',
    'mp_profile_save_changes', 'color_picker_menu_initialize',
    'color_picker_menu_dispose', 'color_picker_select_color',
    'player_prof_begin_editing', 'player_prof_end_editing',
    'player_prof_change_name', 'player_prof_save_changes',
    'player_prof_init_control_settings', 'player_prof_init_adv_ctrl_settings',
    'player_prof_save_control_settings', 'player_prof_save_adv_ctrl_settings',
    'mp_game_player_quit', 'main_menu_switch_to_solo_game',
    'request_del_player_profile', 'request_del_playlist_profile',
    'final_del_player_profile', 'final_del_playlist_profile',
    'cancel_profile_delete',
    'create_and_edit_playlist_profile', 'create_and_edit_player_profile',
    'net_game_speed_start', 'net_game_delay_start',
    'net_server_accept_connection', 'net_server_defer_start',
    'net_server_allow_start', 'disable_if_no_xdemos', 'run_xdemos',
    'sp_reset_controller_choices',
    'sp_set_p1_controller_choices', 'sp_set_p2_controller_choices',
    'error_if_no_network_connection', 'start_server_if_none_advertised',
    'net_game_unjoin_player', 'close_if_not_editing_profile',
    'exit_to_xbox_dashboard', 'new_campaign_chosen', 'new_campaign_decision',
    'pop_history_stack_once', 'difficulty_menu_init', 'begin_music_fade_out',
    'new_game_if_no_player_profile', 'exit_gracefully_to_xbox_dashboard',
    'pause_game_invert_pitch', 'start_new_coop_game',
    'pause_game_invert_spinner_set', 'pause_game_invert_spinner_get',
    'main_menu_quit_game',
    'mouse_emit_ACCEPT_event', 'mouse_emit_BACK_event',
    'mouse_emit_DPAD_LEFT_event', 'mouse_emit_DPAD_RIGHT_event',
    'mouse_spinner_3wide_click',
    'controls_screen_init', 'video_screen_init', 'controls_begin_binding',
    'gamespy_screen_init', 'gamespy_screen_dispose',
    'gamespy_select_header', 'gamespy_select_item', 'gamespy_select_button',
    'player_prof_init_mouse_set', 'player_prof_change_mouse_set',
    'player_prof_init_audio_set', 'player_prof_change_audio_set',
    'player_prof_change_video_set',
    'controls_screen_dispose', 'controls_screen_change_set',
    'mouse_emit_X_event', 'gamepad_screen_init', 'gamepad_screen_dispose',
    'gamepad_screen_change_gamepads', 'gamepad_screen_select_item',
    'mouse_screen_defaults', 'audio_screen_defaults', 'video_screen_defaults',
    'controls_screen_defaults', 'profile_set_edit_begin',
    'profile_manager_delete', 'profile_manager_select',
    'gamespy_dismiiss_error', 'server_settings_init',
    'server_set_edit_server_name', 'server_set_edit_server_password',
    'server_set_start_game', 'video_test_dialog_init',
    'video_test_dialog_dispose', 'video_test_dialog_accept',
    'gamespy_dismiss_filters', 'gamespy_update_filter_settings',
    'gamespy_dismiss_back_handler', 'mouse_spinner_1wide_click',
    'controls_back_handler', 'controls_advanced_launch', 'controls_advanced_ok',
    'mp_pause_menu_open', 'mp_game_options_open', 'mp_choose_team',
    'mp_prof_init_vehicle_options', 'mp_prof_save_vehicle_options',
    'single_prev_cl_item_active',
    'mp_prof_init_teamplay_options', 'mp_prof_save_teamplay_options',
    'mp_game_options_choose', 'emit_custom_activation_event',
    'player_prof_cancel_audio_set', 'player_prof_init_network_options',
    'player_prof_save_network_options', 'credits_post_render',
    'difficulty_item_select', 'credits_initialize', 'credits_dispose',
    'gamespy_get_patch', 'video_screen_dispose',
    'campaign_menu_init', 'campaign_menu_continue',
    'load_game_menu_init', 'load_game_menu_dispose', 'load_game_menu_activated',
    'solo_menu_save_checkpoint', 'mp_type_set_mode',
    'checking_for_updates_ok', 'checking_for_updates_dismiss',
    'direct_ip_connect_init', 'direct_ip_connect_go', 'direct_ip_edit_field',
    'network_settings_edit_a_port', 'network_settings_defaults',
    'load_game_menu_delete_request', 'load_game_menu_delete_finish'
    )

widget_bounds = QStruct("",
    SInt16("t"), SInt16("l"), SInt16("b"),  SInt16("r"),
    ORIENT='h', SIZE=8
    )

game_data_input = Struct("game_data_input",
    SEnum16("function", *game_data_input_functions),
    SIZE=36
    )

event_handler = Struct("event_handler",
    Bool32('flags',
        "close_current_widget",
        "close_other_widget",
        "close_all_widgets",
        "open_widget",
        "reload_self",
        "reload_other_widget",
        "give_focus_to_widget",
        "run_function",
        "replace_self_with_widget",
        "go_back_to_previous_widget",
        "run_scenario_script",
        "try_to_branch_on_failure",
        ),
    SEnum16("event_type", *event_types),
    SEnum16("function", *event_functions),
    dependency("widget_tag", "DeLa"),
    dependency("sound_effect", "snd!"),
    ascii_str32("script"),
    SIZE=72
    )

s_and_r_reference = Struct("search_and_replace_reference",
    ascii_str32("search_string"),
    SEnum16("replace_function",
        "NULL",
        "widgets_controller",
        "build_number",
        "pid",
        ),
    SIZE=34
    )

conditional_widget = Struct("conditional_widget",
    dependency("widget_tag", "DeLa"),
    ascii_str32("name"),  # UNUSED
    Bool32("flags",
        "load_if_event_handler_function_fails",
        ),
    SInt16("custom_controller_index"),  # UNUSED
    SIZE=80
    )

child_widget = Struct("child_widget",
    dependency("widget_tag", "DeLa"),
    ascii_str32("name"),  # UNUSED
    Bool32("flags",
        "use_custom_controller_index",
        ),
    SInt16("custom_controller_index"),
    SInt16("vertical_offset"),
    SInt16("horizontal_offset"),
    SIZE=80
    )

DeLa_body = Struct("tagdata",
    SEnum16("widget_type",
        "container",
        "text_box",
        "spinner_list",
        "column_list",
        "game_model",  # not implemented
        "movie",  # not implemented
        "custom"  # not implemented
        ),
    SEnum16("controller_index",
        "player_1",
        "player_2",
        "player_3",
        "player_4",
        "any_player"
        ),
    ascii_str32("name"),
    QStruct("bounds", INCLUDE=widget_bounds),
    Bool32('flags',
        "pass_unhandled_events_to_focused_child",
        "pause_game_time",
        "flash_background_bitmap",
        "dpad_up_down_tabs_thru_children",
        "dpad_left_right_tabs_thru_children",
        "dpad_up_down_tabs_thru_list_items",
        "dpad_left_right_tabs_thru_list_items",
        "dont_focus_a_specific_child_widget",
        "pass_unhandled_events_to_all_children",
        "return_to_main_menu_if_no_history",
        "always_use_tag_controller_index",
        "always_use_nifty_render_fx",
        "dont_push_history",
        "force_handle_mouse"
        ),
    SInt32("auto_close_time", SIDETIP="milliseconds"),
    SInt32("auto_close_fade_time", SIDETIP="milliseconds"),
    dependency("background_bitmap", "bitm"),

    reflexive("game_data_inputs", game_data_input, 64),
    reflexive("event_handlers", event_handler, 32),
    reflexive("search_and_replace_references",
              s_and_r_reference, 32, DYN_NAME_PATH='.search_string'),

    Pad(128),
    Struct("text_box",
        dependency("text_label_unicode_strings_list", "ustr"),
        dependency("text_font", "font"),
        QStruct("text_color", INCLUDE=argb_float),
        SEnum16("justification",
            "left",
            "right",
            "center",
            ),
        # as weird as it sounds, these flags are off alignment by 2
        Bool32("flags",
            "editable",
            "password",
            "flashing",
            "dont_do_that_weird_focus_test",
            ),
        BytesRaw("unknown2", SIZE=10, VISIBLE=False),

        FlSInt16("unknown3", VISIBLE=False),
        SInt16("string_list_index"),
        SInt16("horizontal_offset"),
        SInt16("vertical_offset")
        ),

    Pad(28),
    Struct("list_items",
        Bool32("flags",
            "list_items_generated_in_code",
            "list_items_from_string_list_tag",
            "list_items_only_one_tooltip",
            "list_single_preview_no_scroll"
            )
        ),

    Struct("spinner_list",
        dependency("list_header_bitmap", "bitm"),
        dependency("list_footer_bitmap", "bitm"),
        QStruct("header_bounds", INCLUDE=widget_bounds),
        QStruct("footer_bounds", INCLUDE=widget_bounds)
        ),

    Pad(32),
    Struct("column_list",
        dependency("extended_description_widget", "DeLa")
        ),

    Pad(288),
    reflexive("conditional_widgets", conditional_widget, 32,
        DYN_NAME_PATH='.widget_tag.filepath'),

    Pad(256),
    reflexive("child_widgets", child_widget, 32,
        DYN_NAME_PATH='.widget_tag.filepath'),

    SIZE=1004
    )

def get():
    return DeLa_def

DeLa_def = TagDef("DeLa",
    blam_header('DeLa'),
    DeLa_body,

    ext=".ui_widget_definition", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.h2.defs import __all__ as all_def_names
from reclaimer.hek.handler import HaloHandler
from reclaimer.data_extraction import h2_data_extractors

class Halo2Handler(HaloHandler):
    frozen_imp_paths = all_def_names
    tag_header_engine_id = "b2am"
    default_defs_path = "reclaimer.h2.defs"

    tag_data_extractors = h2_data_extractors



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from struct import unpack

from supyr_struct.field_type_methods import *
from reclaimer.constants import *


def tag_cstring_parser(self, desc, node=None, parent=None, attr_index=None,
                       rawdata=None, root_offset=0, offset=0, **kwargs):
    """
    """
    assert parent is not None and attr_index is not None, (
        "parent and attr_index must be provided " +
        "and not None when reading a data field.")

    if rawdata is not None:
        max_len = desc.get(MAX, 0xFFffFFff)
        offset = parent.get_meta(POINTER, attr_index, **kwargs)

        start = root_offset + offset
        charsize = self.size
        delimiter = self.delimiter

        # if the character size is greater than 1 we need to do special
        # checks to ensure the position the null terminator was found at
        # is not overlapping the boundary between individual characters.
        size = rawdata.find(delimiter, start) - start
        if size > max_len:
            size = max_len

        if size >= 0:
            rawdata.seek(start)
            # read and store the string
            parent[attr_index] = self.decoder(
                rawdata.read(size), desc=desc,
                parent=parent, attr_index=attr_index)

            return offset + size + charsize

    parent[attr_index] = desc.get(DEFAULT, self.default())

    return offset


def tag_ref_str_sizecalc(self, node, **kwargs):
    '''
    Used to calculate the size of a tag reference string from a given string
    '''
    node = node.split(self.str_delimiter)[0]
    if node:
        return len(node) + 1
    return 0


def get_set_zone_asset_size(node=None, parent=None, attr_index=None,
                            rawdata=None, new_value=None, **kwargs):
    if new_value is not None:
        parent.size = new_value
    elif "map_pointer_converter" in kwargs:
        return 0
    return parent.size


def tag_ref_str_size(node=None, parent=None, attr_index=None,
                     rawdata=None, new_value=None, **kwargs):
    '''Used to retrieve or set the byte size of a Halo tag
    reference string. If the string is empty, the actual amount
    of bytes it takes up is zero, otherwise it is (1+length) bytes.
    This is to account for the delimiter.

    When setting the size, the provided new_value is expected to
    be including the delimiter, so the reverse operation is applied.
    If the string's length is 1(only a delimiter), the bytes size
    is zero, but otherwise it is (length-1).

    Lengths of 1 cant exist.'''

    if new_value is None:
        strlen = parent.path_length
        return strlen + bool(strlen)

    parent.path_length = (new_value - 1)*(new_value > 1)


def encode_tag_ref_str(self, node, parent=None, attr_index=None):
    """This function is the same as encode_string, except that
    when a halo reference string has zero length, the string doesnt
    actually exist. It's not just a delimiter character, the string
    isn't stored at all. To make it work, we instead return an
    empty bytes object if the string length is zero"""
    if node:
        return encode_string(self, node, parent=parent, attr_index=attr_index)
    return b''


def tag_ref_str_parser(self, desc, node=None, parent=None, attr_index=None,
                       rawdata=None, root_offset=0, offset=0, **kwargs):
    """
    """
    assert parent is not None and attr_index is not None, (
        "parent and attr_index must be provided " +
        "and not None when reading a data field.")
    tag_index_manager = kwargs.get("tag_index_manager")
    if tag_index_manager:
        tag_index_ref = tag_index_manager.get_tag_index_ref(parent.id)
        if tag_index_ref is None or (not kwargs.get("indexed") and
                                     tag_index_ref.id != parent.id):
            parent[attr_index] = ""
        else:
            parent[attr_index] = tag_index_ref.path

    elif rawdata:
        # read and store the node
        rawdata.seek(root_offset + offset)
        size = parent.get_size(attr_index, root_offset=root_offset,
                               offset=offset, rawdata=rawdata, **kwargs)
        parent[attr_index] = self.decoder(rawdata.read(size), desc=desc,
                                          parent=parent, attr_index=attr_index)
        return offset + size
    else:
        parent[attr_index] = desc.get(DEFAULT, self.default())

    return offset


def read_string_id_string(parent=None, attr_index=None, rawdata=None, offset=0,
                          map_string_id_manager=None, **kwargs):
    assert parent is not None
    if map_string_id_manager:
        parent[attr_index] = map_string_id_manager.get_string(parent)
    elif "map_pointer_converter" not in kwargs:
        desc = parent.desc
        str_len = (parent[0] & 0xFFffFFff) >> (desc[STRINGID_IDX_BITS] +
                                               desc[STRINGID_SET_BITS])
        if str_len:
            rawdata.seek(offset)
            parent[attr_index] = rawdata.read(str_len).strip(b'\x00').decode(
                encoding="latin-1")
        else:
            parent[attr_index] = ""

        offset += str_len + 1  # add 1 for the null terminator
    else:
        parent[attr_index] = ""

    return offset


def write_string_id_string(parent=None, attr_index=None,
                           writebuffer=None, offset=0, **kwargs):
    assert parent is not None
    if "map_pointer_converter" in kwargs:
        return offset

    raw_string = parent.string.encode(encoding="latin-1") + b'\x00'
    writebuffer.seek(offset)
    writebuffer.write(raw_string)
    return offset + len(raw_string)


def get_set_string_id_size(parent=None, attr_index=None,
                           new_value=None, **kwargs):
    desc = parent.desc
    string_id = parent[0]
    idx_set_bit_ct = desc[STRINGID_IDX_BITS] + desc[STRINGID_SET_BITS]
    len_bit_ct = desc[STRINGID_LEN_BITS]
    if new_value is None:
        return (string_id & 0xFFffFFff) >> idx_set_bit_ct
    elif new_value > (1 << len_bit_ct):
        raise ValueError("String ID's cannot be longer than %s characters." %
                         ((1 << len_bit_ct) - 1))
    elif new_value > 0:
        new_value -= 1  # subtract the null delimiter
    else:
        assert new_value == 0
    parent[0] = (string_id & ((1 << idx_set_bit_ct) - 1)) + (
        new_value << idx_set_bit_ct)


def reflexive_parser(self, desc, node=None, parent=None, attr_index=None,
                     rawdata=None, root_offset=0, offset=0, **kwargs):
    """
    """
    try:
        __lsi__ = list.__setitem__
        orig_offset = offset
        if node is None:
            parent[attr_index] = node = desc.get(NODE_CLS, self.node_cls)\
                (desc, parent=parent)

        # If there is rawdata to build the structure from
        if rawdata is not None:
            struct_off = root_offset + offset

            if self.f_endian == '=':
                for i, off in enumerate(desc[ATTR_OFFS]):
                    off += struct_off
                    typ = desc[i][TYPE]
                    __lsi__(node, i, typ.struct_unpacker(
                        rawdata[off:off + typ.size])[0])
            elif self.f_endian == '<':
                for i, off in enumerate(desc[ATTR_OFFS]):
                    off += struct_off
                    typ = desc[i][TYPE].little
                    __lsi__(node, i, typ.struct_unpacker(
                        rawdata[off:off + typ.size])[0])
            else:
                for i, off in enumerate(desc[ATTR_OFFS]):
                    off += struct_off
                    typ = desc[i][TYPE].big
                    __lsi__(node, i, typ.struct_unpacker(
                        rawdata[off:off + typ.size])[0])

            # increment offset by the size of the struct
            offset += desc[SIZE]
        else:
            for i in range(len(node)):
                __lsi__(node, i, desc[i].get(
                    DEFAULT, desc[i][TYPE].default()))

        s_desc = desc.get(STEPTREE)
        if s_desc:
            pointer_converter = kwargs.get('map_pointer_converter')
            if pointer_converter is not None:
                file_ptr = pointer_converter.v_ptr_to_f_ptr(node[1])
                if kwargs.get("safe_mode") and not s_desc.get(IGNORE_SAFE_MODE):
                    # make sure the reflexive sizes are less than or equal to
                    # the max number of entries allowed in the reflexive
                    node[0] = max(0, min(node[0], s_desc.get(MAX, node[0])))

                if (file_ptr < 0 or file_ptr +
                    node[0]*s_desc[SUB_STRUCT].get(SIZE, 0) > len(rawdata)):
                    # the reflexive is corrupt for some reason
                    #    (ex: bad hek+ extraction)
                    node[0] = node[1] = 0

            if not node[0]:
                # reflexive is empty. no need to provide rawdata
                s_desc[TYPE].parser(s_desc, None, node, STEPTREE, None)
            elif pointer_converter is not None:
                # parsing tag from a map
                s_desc[TYPE].parser(s_desc, None, node, STEPTREE, rawdata,
                                      root_offset, file_ptr, **kwargs)
            elif 'steptree_parents' not in kwargs:
                offset = s_desc[TYPE].parser(s_desc, None, node, STEPTREE,
                                               rawdata, root_offset, offset,
                                               **kwargs)
            else:
                kwargs['steptree_parents'].append(node)

        # pass the incremented offset to the caller
        return offset
    except Exception as e:
        # if the error occurred while parsing something that doesnt have an
        # error report routine built into the function, do it for it.
        kwargs.update(buffer=rawdata, root_offset=root_offset)
        if 's_desc' in locals():
            e = format_parse_error(e, field_type=s_desc.get(TYPE), desc=s_desc,
                                  parent=node, attr_index=STEPTREE,
                                  offset=offset, **kwargs)
        elif 'i' in locals():
            e = format_parse_error(e, field_type=desc[i].get(TYPE),
                                   desc=desc[i], parent=node, attr_index=i,
                                   offset=offset, **kwargs)
        e = format_parse_error(e, field_type=self, desc=desc,
                               parent=parent, attr_index=attr_index,
                               offset=orig_offset, **kwargs)
        raise e


def rawdata_ref_parser(self, desc, node=None, parent=None, attr_index=None,
                       rawdata=None, root_offset=0, offset=0, **kwargs):
    try:
        orig_offset = offset
        if node is None:
            parent[attr_index] = node = desc.get(NODE_CLS, self.node_cls)\
                (desc, parent=parent, init_attrs=rawdata is None)

        # If there is rawdata to build the structure from
        if rawdata is not None:
            # loop once for each field in the node
            for i, off in enumerate(desc['ATTR_OFFS']):
                desc[i][TYPE].parser(desc[i], None, node, i, rawdata,
                                     root_offset, offset + off, **kwargs)

            # increment offset by the size of the struct
            offset += desc[SIZE]

        s_desc = desc.get(STEPTREE)
        if s_desc:
            pointer_converter = kwargs.get("map_pointer_converter")

            if pointer_converter is not None:
                if kwargs.get("safe_mode") and not s_desc.get(IGNORE_SAFE_MODE):
                    node[0] = max(0, min(node[0], s_desc.get(MAX, node[0])))

            if kwargs.get("parsing_resource"):
                # parsing JUST metadata from a resource cache
                node_size = node[0]
                if 'steptree_parents' not in kwargs and pointer_converter is not None:
                    # need to skip over the rawdata
                    offset += node_size + node[2]

                node[0] = 0
                s_desc[TYPE].parser(s_desc, None, node, STEPTREE, None)
                node[0] = node_size
            elif pointer_converter is not None:
                file_ptr = pointer_converter.v_ptr_to_f_ptr(node[3])
                if not node[3] or file_ptr < 0:
                    file_ptr = node[2]

                if not node[0] or (file_ptr + node[0] > len(rawdata) or
                                   file_ptr <= 0):
                    # data is stored in a resource map, or the size is invalid
                    s_desc[TYPE].parser(s_desc, None, node, STEPTREE, None)
                else:
                    s_desc[TYPE].parser(
                        s_desc, None, node, STEPTREE, rawdata,
                        root_offset, file_ptr, **kwargs)
            elif 'steptree_parents' not in kwargs:
                offset = s_desc[TYPE].parser(s_desc, None, node, STEPTREE,
                                               rawdata, root_offset, offset,
                                               **kwargs)
            else:
                kwargs['steptree_parents'].append(node)

        # pass the incremented offset to the caller
        return offset
    except Exception as e:
        # if the error occurred while parsing something that doesnt have an
        # error report routine built into the function, do it for it.
        kwargs.update(buffer=rawdata, root_offset=root_offset)
        if 's_desc' in locals():
            e = format_parse_error(e, field_type=s_desc.get(TYPE), desc=s_desc,
                                  parent=node, attr_index=STEPTREE,
                                  offset=offset, **kwargs)
        elif 'i' in locals():
            e = format_parse_error(e, field_type=desc[i].get(TYPE),
                                   desc=desc[i], parent=node, attr_index=i,
                                   offset=offset, **kwargs)
        e = format_parse_error(e, field_type=self, desc=desc,
                               parent=parent, attr_index=attr_index,
                               offset=orig_offset, **kwargs)
        raise e


def tag_ref_str_serializer(self, node, parent=None, attr_index=None,
                           writebuffer=None, root_offset=0, offset=0, **kwargs):
    if "map_pointer_converter" in kwargs:
        # don't serialize the string
        return offset

    node_bytes = self.encoder(node, parent, attr_index)
    writebuffer.seek(root_offset + offset)
    writebuffer.write(node_bytes)
    size = parent.get_size(attr_index, root_offset=root_offset,
                           offset=offset, **kwargs)
    if size - len(node_bytes):
        writebuffer.write(b'\x00'*(size - len(node_bytes)))
    return offset + size



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = ( 'JmsModel', )


import math

from ..constants import JMS_PERM_CANNOT_BE_RANDOMLY_CHOSEN_TOKEN


class JmsModel:
    name = ""

    version = ""

    perm_name = "__base"
    lod_level = "superhigh"
    is_random_perm = True

    node_list_checksum = 0
    nodes = ()
    materials = ()
    regions = ()
    markers = ()
    verts   = ()
    tris    = ()

    def __init__(self, name="", node_list_checksum=0, nodes=None,
                 materials=None, markers=None, regions=None,
                 verts=None, tris=None, version="8200"):

        name = name.strip(" ")
        perm_name = name
        if name.startswith(JMS_PERM_CANNOT_BE_RANDOMLY_CHOSEN_TOKEN):
            self.is_random_perm = False
            perm_name = perm_name.lstrip(JMS_PERM_CANNOT_BE_RANDOMLY_CHOSEN_TOKEN)

        self.lod_level = "superhigh"
        for lod_level in ("superhigh", "high", "medium", "superlow", "low"):
            if perm_name.lower().endswith(lod_level):
                perm_name = perm_name[: -len(lod_level)].strip(" ")
                self.lod_level = lod_level
                break

        node_list_checksum = node_list_checksum

        self.name = name
        self.version = version
        self.perm_name = perm_name
        self.node_list_checksum = node_list_checksum
        self.nodes = nodes if nodes else []
        self.materials = materials if materials else []
        self.regions = regions if regions else ["__unnamed"]
        self.markers = markers if markers else []
        self.verts   = verts   if verts   else []
        self.tris    = tris    if tris    else []

    def calculate_vertex_normals(self):
        verts = self.verts
        vert_ct = len(verts)
        sqrt = math.sqrt

        v_indices = (0, 1, 2)
        binormals = [[0, 0, 0, 0] for i in range(vert_ct)]
        tangents  = [[0, 0, 0, 0] for i in range(vert_ct)]
        for tri in self.tris:
            for tri_i in v_indices:
                v_i = tri[tri_i]
                if v_i >= vert_ct:
                    continue

                v0 = verts[v_i]
                v1 = verts[tri[(tri_i + 1) % 3]]
                v2 = verts[tri[(tri_i + 2) % 3]]
                b = binormals[v_i]
                t = tangents[v_i]

                x0 = v1.pos_x - v0.pos_x;
                x1 = v2.pos_x - v0.pos_x;
                y0 = v1.pos_y - v0.pos_y;
                y1 = v2.pos_y - v0.pos_y;
                z0 = v1.pos_z - v0.pos_z;
                z1 = v2.pos_z - v0.pos_z;


                s0 = v1.tex_u - v0.tex_u;
                s1 = v2.tex_u - v0.tex_u;
                t0 = v1.tex_v - v0.tex_v;
                t1 = v2.tex_v - v0.tex_v;

                r = s0 * t1 - s1 * t0
                if r == 0:
                    continue

                r = 1 / r

                bi = -(s0 * x1 - s1 * x0) * r
                bj = -(s0 * y1 - s1 * y0) * r
                bk = -(s0 * z1 - s1 * z0) * r
                b_len = sqrt(bi**2 + bj**2 + bk**2)

                ti = (t1 * x0 - t0 * x1) * r
                tj = (t1 * y0 - t0 * y1) * r
                tk = (t1 * z0 - t0 * z1) * r
                t_len = sqrt(ti**2 + tj**2 + tk**2)

                if b_len:
                    b[0] += bi / b_len
                    b[1] += bj / b_len
                    b[2] += bk / b_len
                    b[3] += 1

                if t_len:
                    t[0] += ti / t_len
                    t[1] += tj / t_len
                    t[2] += tk / t_len
                    t[3] += 1

        for i in range(vert_ct):
            vert = verts[i]
            b = binormals[i]
            t = tangents[i]

            if b[3]:
                vert.binorm_i = b[0] / b[3]
                vert.binorm_j = b[1] / b[3]
                vert.binorm_k = b[2] / b[3]

            if t[3]:
                vert.tangent_i = t[0] / t[3]
                vert.tangent_j = t[1] / t[3]
                vert.tangent_k = t[2] / t[3]

    def optimize_geometry(self, exact_compare=True):
        verts = self.verts
        vert_ct = len(verts)

        # this will map the verts to prune to the vert they are identical to
        dup_vert_map = {}
        similar_vert_map = {}

        if exact_compare:
            for i in range(len(verts)):
                v = verts[i]
                similar_vert_map.setdefault(
                    (v.pos_x, v.pos_y, v.pos_z), []).append(i)

            # loop over all verts and figure out which ones to replace with others
            for similar_vert_indices in similar_vert_map.values():
                for i in range(len(similar_vert_indices) - 1):
                    orig_idx = similar_vert_indices[i]
                    if orig_idx in dup_vert_map:
                        continue

                    vert_a = verts[orig_idx]
                    vert_a_i = vert_a.norm_i; vert_a_j = vert_a.norm_j; vert_a_k = vert_a.norm_k
                    vert_a_u = vert_a.tex_u;  vert_a_v = vert_a.tex_v
                    vert_a_n0 = vert_a.node_0; vert_a_n1 = vert_a.node_1
                    vert_a_n1w = vert_a.node_1_weight
                    for j in similar_vert_indices[i + 1: ]:
                        if j in dup_vert_map:
                            continue

                        vert_b = verts[j]
                        if (vert_a_i != vert_b.norm_i or
                            vert_a_j != vert_b.norm_j or
                            vert_a_k != vert_b.norm_k):
                            continue
                        elif vert_a_n1w != vert_b.node_1_weight:
                            continue
                        elif vert_a_n0 != vert_b.node_0 or vert_a_n1 != vert_b.node_1:
                            continue
                        elif vert_a_u != vert_b.tex_u or vert_a_v != vert_b.tex_v:
                            continue

                        dup_vert_map[j] = orig_idx
        else:
            for i in range(len(verts)):
                v = verts[i]
                similar_vert_map.setdefault(
                    (round(v.pos_x + 0.001, 3),
                     round(v.pos_y + 0.001, 3),
                     round(v.pos_z + 0.001, 3)), []).append(i)

            # loop over all verts and figure out which ones to replace with others
            for similar_vert_indices in similar_vert_map.values():
                for i in range(len(similar_vert_indices) - 1):
                    orig_idx = similar_vert_indices[i]
                    if orig_idx in dup_vert_map:
                        continue

                    vert_a = verts[orig_idx]
                    for j in similar_vert_indices[i + 1: ]:
                        if j not in dup_vert_map and verts[j] == vert_a:
                            dup_vert_map[j] = orig_idx

        if not dup_vert_map:
            # nothing to optimize away
            return

        # remap any duplicate triangle vert indices to the original
        get_mapped_vert = dup_vert_map.get
        for tri in self.tris:
            tri.v0 = get_mapped_vert(tri.v0, tri.v0)
            tri.v1 = get_mapped_vert(tri.v1, tri.v1)
            tri.v2 = get_mapped_vert(tri.v2, tri.v2)

        # copy the verts list so we can modify it without side-effects
        new_vert_ct = vert_ct - len(dup_vert_map)
        self.verts = new_verts = self.verts[: new_vert_ct]

        shift_map = {}
        copy_idx = vert_ct - 1
        # loop over all duplicate vert indices and move any vertices
        # on the high end of the vert list down to fill in the empty
        # spaces left by the duplicate verts we're removing.
        for dup_i in sorted(dup_vert_map):
            while copy_idx in dup_vert_map:
                # keep looping until we get to a vert we can move
                # from its high index to overwrite the low index dup
                copy_idx -= 1

            if copy_idx <= dup_i or dup_i >= new_vert_ct:
                # cant copy any lower. all upper index verts are duplicates
                break

            # move the vert from its high index to the low index dup
            new_verts[dup_i] = verts[copy_idx]
            shift_map[copy_idx] = dup_i
            copy_idx -= 1

        # remap any triangle vert indices
        get_mapped_vert = shift_map.get
        for tri in self.tris:
            tri.v0 = get_mapped_vert(tri.v0, tri.v0)
            tri.v1 = get_mapped_vert(tri.v1, tri.v1)
            tri.v2 = get_mapped_vert(tri.v2, tri.v2)

    def get_node_depths(self):
        node_depths = [-1] * len(self.nodes)
        if not self.nodes:
            return node_depths

        node_depths[0] = 0

        seen_hierarchy = set((-1, ))
        # figure out the hierarchy depth of each node
        for i in range(len(self.nodes)):
            child_depth = node_depths[i] + 1
            child_idx = self.nodes[i].first_child
            while (child_idx not in seen_hierarchy and
                   child_idx in range(len(self.nodes))):
                seen_hierarchy.add(child_idx)
                node_depths[child_idx] = child_depth
                child_idx = self.nodes[child_idx].sibling_index

        return node_depths

    def verify_nodes_valid(self):
        errors = []
        if len(self.nodes) == 0:
            errors.append("No nodes. Must contain at least one node.")
            return errors

        if len(self.nodes) >= 64:
            errors.append("Too many nodes. Max count is 64.")

        seen_names = set()
        for i in range(len(self.nodes)):
            n = self.nodes[i]
            if n.first_child >= len(self.nodes):
                errors.append("First child of node '%s' is invalid." % n.name)
            elif n.sibling_index >= len(self.nodes):
                errors.append("Sibling node of node '%s' is invalid." % n.name)
            elif len(n.name) >= 32:
                errors.append("Node name node '%s' is too long." % n.name)
            elif n.name.lower() in seen_names:
                errors.append("Multiple nodes named '%s'." % n.name)

            seen_names.add(n.name.lower())

        if self.nodes and self.nodes[0].sibling_index != -1:
            errors.append("Root node must not have siblings.")

        node_depths = self.get_node_depths()

        # make sure the nodes are sorted in increasing hierarchy depth
        curr_depth = node_depths[0]
        prev_name = ""
        for i in range(len(node_depths)):
            if curr_depth > node_depths[i]:
                errors.append("Nodes are not sorted by hierarchy depth.")
                break

            curr_name = self.nodes[i].name
            if curr_depth != node_depths[i]:
                curr_depth = node_depths[i]
            elif curr_name < prev_name:
                errors.append(("Nodes within depth %s are not sorted "
                               "alphabetically.") % curr_depth)
                break

            prev_name = curr_name

        sib_errors = set()
        child_errors = set()
        seen_hierarchy = set()
        for node in self.nodes:
            sib_idx = node.sibling_index
            child_idx = node.first_child

            if child_idx in child_errors:
                pass
            elif child_idx in seen_hierarchy:
                errors.append(
                    "Node %s is specified as the child of multiple nodes." %
                    self.nodes[child_idx].name)
            elif child_idx >= 0:
                seen_hierarchy.add(child_idx)
                seen_children = set()
                while child_idx >= 0:
                    child_node = self.nodes[child_idx]
                    if child_idx in seen_children:
                        errors.append("Circular reference in children " +
                                      "of node '%s'." % node.name)
                        break

                    seen_children.add(child_idx)
                    child_idx = child_node.first_child

            if sib_idx in sib_errors:
                pass
            elif sib_idx in seen_hierarchy:
                errors.append(
                    "Node %s is specified as the sibling of multiple nodes." %
                    self.nodes[sib_idx].name)
            elif sib_idx >= 0:
                seen_hierarchy.add(sib_idx)
                seen_siblings = set()
                while sib_idx >= 0:
                    sib_node = self.nodes[sib_idx]
                    if sib_idx in seen_siblings:
                        errors.append("Circular reference in siblings " +
                                      "of node '%s'." % node.name)
                        break

                    seen_siblings.add(sib_idx)
                    sib_idx = sib_node.sibling_index


        return errors

    def verify_models_match(self, other_jms):
        errors = list(other_jms.verify_jms())
        if len(other_jms.nodes) != len(self.nodes):
            errors.append("Node counts do not match.")
            return errors

        for i in range(len(self.nodes)):
            if self.nodes[i] != other_jms.nodes[i]:
                errors.append("Nodes '%s' do not match." % i)

        return errors

    def verify_jms(self):
        crc = self.node_list_checksum
        mats  = self.materials
        markers = self.markers
        regions = self.regions

        if isinstance(self, JmsModel):
            perm_meshes = {self.name: self}
        else:
            perm_meshes = self.perm_meshes

        node_error = False

        node_ct = len(self.nodes)
        region_ct = len(regions)
        mat_ct = len(mats)

        errors = self.verify_nodes_valid()
        if errors:
            return errors

        err_str = "Invalid %s index in %s(s)."
        for region_name in regions:
            if len(region_name) >= 32:
                errors.append("Region name '%s' is too long." % region_name)

        for marker in markers:
            if marker.parent >= node_ct:
                errors.append(err_str % ("parent", "marker"))
            elif marker.region >= region_ct:
                errors.append(err_str % ("region", "marker"))
            elif len(marker.name) >= 32:
                errors.append("Marker name '%s' is too long." % marker.name)
            else:
                continue
            break

        for perm_name in sorted(perm_meshes):
            verts = perm_meshes[perm_name].verts
            tris  = perm_meshes[perm_name].tris
            vert_ct = len(verts)
            for tri in tris:
                if (tri.v0 < 0 or tri.v1 < 0 or tri.v2 < 0 or
                    tri.v0 >= vert_ct or tri.v1 >= vert_ct or tri.v2 >= vert_ct):
                    errors.append(err_str % ("vertex", "triangle"))
                elif tri.region >= region_ct:
                    errors.append(err_str % ("region", "triangle"))
                elif tri.shader >= mat_ct:
                    errors.append(err_str % ("shader", "triangle"))
                else:
                    continue
                break

            for vert in verts:
                if vert.node_0 >= node_ct:
                    errors.append(err_str % ("node_0", "vertex"))
                elif vert.node_1 >= node_ct:
                    errors.append(err_str % ("node_1", "vertex"))
                else:
                    continue
                break

        return errors



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: -DeToX-
# 	Found Raw Identifier
# revision: 3		author: Lord Zedd
# 	Lookin Nice
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

jmad_animation_desired_compression = (
    "best_score",
    "best_compression",
    "best_accuracy",
    "best_fullframe",
    "best_small_keyframe",
    "best_large_keyframe",
    )

jmad_animation_frame_event_type = (
    "primary_keyframe",
    "secondary_keyframe",
    "left_foot",
    "right_foot",
    "allow_interruption",
    "transition_a",
    "transition_b",
    "transition_c",
    "transition_d",
    "both_feet_shuffle",
    "body_impact",
    )

jmad_animation_frame_info_type = (
    "none",
    "dx_dy",
    "dx_dy_dyaw",
    "dx_dy_dz_dyaw",
    )

jmad_animation_type = (
    "base",
    "overlay",
    "replacement",
    )

jmad_leg_anchors = (
    "false",
    "true",
    )

jmad_mode_weapon_clas_weapon_type_transition_destination_frame_event_link = (
    "no_keyframe",
    "keyframe_type_a",
    "keyframe_type_b",
    "keyframe_type_c",
    "keyframe_type_d",
    )

jmad_object_overlay_function_controls = (
    "frame",
    "scale",
    )


jmad_skeleton_node = Struct("skeleton_node",
    h3_string_id("name"),
    SInt16("next_sibling_node_index"),
    SInt16("first_child_node_index"),
    SInt16("parent_node_index"),
    Bool8("model_flags",
        "primary_model",
        "secondary_model",
        "local_root",
        "left_hand",
        "right_hand",
        "left_arm_member",
        ),
    Bool8("node_joint_flags",
        "ball_socket",
        "hinge",
        "no_movement",
        ),
    QStruct("base_vector", INCLUDE=ijk_float),
    Float("vector_range"),
    Float("z_position"),
    ENDIAN=">", SIZE=32
    )


jmad_sound_reference = Struct("sound_reference",
    h3_dependency("sound"),
    Bool16("flags",
        "allow_on_player",
        "left_arm_only",
        "right_arm_only",
        "first_person_only",
        "forward_only",
        "reverse_only",
        ),
    SInt16("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=20
    )


jmad_effect_reference = Struct("effect_reference",
    h3_dependency("effect"),
    Bool16("flags",
        "allow_on_player",
        "left_arm_only",
        "right_arm_only",
        "first_person_only",
        "forward_only",
        "reverse_only",
        ),
    SInt16("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=20
    )


jmad_blend_screen = Struct("blend_screen",
    h3_string_id("label"),
    float_rad("right_yaw_per_frame"),
    float_rad("left_yaw_per_frame"),
    SInt16("right_frame_count"),
    SInt16("left_frame_count"),
    float_rad("down_pitch_per_frame"),
    float_rad("up_pitch_per_frame"),
    SInt16("down_pitch_frame_count"),
    SInt16("up_pitch_frame_count"),
    ENDIAN=">", SIZE=28
    )


jmad_leg = Struct("leg",
    h3_string_id("foot_marker"),
    QStruct("foot", INCLUDE=from_to),
    h3_string_id("ankle_marker"),
    QStruct("ankle", INCLUDE=from_to),
    SEnum16("anchors", *jmad_leg_anchors),
    SInt16("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=28
    )


jmad_animation_frame_event = Struct("frame_event",
    SEnum16("type", *jmad_animation_frame_event_type),
    SInt16("frame"),
    ENDIAN=">", SIZE=4
    )


jmad_animation_sound_event = Struct("sound_event",
    SInt16("sound"),
    SInt16("frame"),
    h3_string_id("marker_name"),
    ENDIAN=">", SIZE=8
    )


jmad_animation_effect_event = Struct("effect_event",
    SInt16("effect"),
    SInt16("frame"),
    h3_string_id("marker_name"),
    ENDIAN=">", SIZE=8
    )


jmad_animation_unknown_2 = Struct("unknown_2",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


jmad_animation_object_space_parent_node = Struct("object_space_parent_node",
    SInt16("node_index"),
    Bool16("component_flags",
        "rotation",
        "translation",
        "scale",
        ),
    QStruct("rotation", INCLUDE=ijkw_sint16),
    QStruct("default_translation", INCLUDE=xyz_float),
    Float("default_scale"),
    ENDIAN=">", SIZE=28
    )


jmad_animation_leg_anchoring_unknown_1 = Struct("unknown_1",
    SInt16("frame_1a"),
    SInt16("frame_2a"),
    SInt16("frame_1b"),
    SInt16("frame_2b"),
    BytesRaw("unknown", SIZE=12, VISIBLE=False),
    ENDIAN=">", SIZE=20
    )


jmad_animation_leg_anchoring = Struct("leg_anchoring",
    SInt16("leg_index"),
    SInt16("unknown_0", VISIBLE=False),
    h3_reflexive("unknown_1", jmad_animation_leg_anchoring_unknown_1),
    ENDIAN=">", SIZE=16
    )


jmad_animation = Struct("animation",
    h3_string_id("name"),
    Float("weight"),
    SInt16("loop_frame_index"),
    Bool16("playback_flags",
        "disable_interpolation_in",
        "disable_interpolation_out",
        "disable_mode_ik",
        "disable_weapon_ik",
        "disable_weapon_aim_first_person",
        "disable_look_screen",
        "disable_transition_adjustment",
        ),
    SInt8("blend_screen"),
    SEnum8("desired_compression", *jmad_animation_desired_compression),
    SEnum8("current_compression", *jmad_animation_desired_compression),
    UInt8("node_count"),
    SInt16("frame_count"),
    SEnum8("type", *jmad_animation_type),
    SEnum8("frame_info_type", *jmad_animation_frame_info_type),
    Bool16("production_flags",
        "do_not_monitor_changes",
        "verify_sound_events",
        "do_not_inherit_for_player_graphs",
        ),
    Bool16("internal_flags",
        ("world_relative", 1 << 1),
        ("compression_disabled", 1 << 5),
        "old_production_checksum",
        "valid_production_checksum",
        ),
    SInt32("node_list_checksum"),
    SInt32("production_checksum"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SInt16("previous_variant_sibling"),
    SInt16("next_variant_sibling"),
    SInt16("raw_information_group_index"),
    SInt16("raw_information_member_index"),
    h3_reflexive("frame_events", jmad_animation_frame_event),
    h3_reflexive("sound_events", jmad_animation_sound_event),
    h3_reflexive("effect_events", jmad_animation_effect_event),
    h3_reflexive("unknown_2", jmad_animation_unknown_2),
    h3_reflexive("object_space_parent_nodes", jmad_animation_object_space_parent_node),
    h3_reflexive("leg_anchoring", jmad_animation_leg_anchoring),
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=5, VISIBLE=False),
    ENDIAN=">", SIZE=136
    )


jmad_mode_weapon_clas_weapon_type_action = Struct("action",
    h3_string_id("label"),
    SInt16("graph_index"),
    SInt16("animation"),
    ENDIAN=">", SIZE=8
    )


jmad_mode_weapon_clas_weapon_type_overlay = Struct("overlay",
    h3_string_id("label"),
    SInt16("graph_index"),
    SInt16("animation"),
    ENDIAN=">", SIZE=8
    )


jmad_mode_weapon_clas_weapon_type_death_and_damage_direction_region = Struct("region",
    SInt16("graph_index"),
    SInt16("animation"),
    ENDIAN=">", SIZE=4
    )


jmad_mode_weapon_clas_weapon_type_death_and_damage_direction = Struct("direction",
    h3_reflexive("regions", jmad_mode_weapon_clas_weapon_type_death_and_damage_direction_region),
    ENDIAN=">", SIZE=12
    )


jmad_mode_weapon_clas_weapon_type_death_and_damage = Struct("death_and_damage",
    h3_string_id("label"),
    h3_reflexive("directions", jmad_mode_weapon_clas_weapon_type_death_and_damage_direction),
    ENDIAN=">", SIZE=16
    )


jmad_mode_weapon_clas_weapon_type_transition_destination = Struct("destination",
    h3_string_id("full_name"),
    h3_string_id("mode_name"),
    h3_string_id("state_name"),
    SEnum8("frame_event_link", *jmad_mode_weapon_clas_weapon_type_transition_destination_frame_event_link),
    SInt8("unknown", VISIBLE=False),
    SInt8("index_a"),
    SInt8("index_b"),
    SInt16("graph_index"),
    SInt16("animation"),
    ENDIAN=">", SIZE=20
    )


jmad_mode_weapon_clas_weapon_type_transition = Struct("transition",
    h3_string_id("full_name"),
    h3_string_id("state_name"),
    SInt16("unknown", VISIBLE=False),
    SInt8("index_a"),
    SInt8("index_b"),
    h3_reflexive("destinations", jmad_mode_weapon_clas_weapon_type_transition_destination),
    ENDIAN=">", SIZE=24
    )


jmad_mode_weapon_clas_weapon_type = Struct("weapon_type",
    h3_string_id("label"),
    h3_reflexive("actions", jmad_mode_weapon_clas_weapon_type_action),
    h3_reflexive("overlays", jmad_mode_weapon_clas_weapon_type_overlay),
    h3_reflexive("death_and_damage", jmad_mode_weapon_clas_weapon_type_death_and_damage),
    h3_reflexive("transitions", jmad_mode_weapon_clas_weapon_type_transition),
    ENDIAN=">", SIZE=52
    )


jmad_mode_weapon_clas_weapon_ik = Struct("weapon_ik",
    h3_string_id("marker"),
    h3_string_id("attach_to_marker"),
    ENDIAN=">", SIZE=8
    )


jmad_mode_weapon_clas = Struct("weapon_clas",
    h3_string_id("label"),
    h3_reflexive("weapon_type", jmad_mode_weapon_clas_weapon_type),
    h3_reflexive("weapon_ik", jmad_mode_weapon_clas_weapon_ik),
    ENDIAN=">", SIZE=28
    )


jmad_mode_mode_ik = Struct("mode_ik",
    h3_string_id("marker"),
    h3_string_id("attach_to_marker"),
    ENDIAN=">", SIZE=8
    )


jmad_mode = Struct("mode",
    h3_string_id("label"),
    h3_reflexive("weapon_class", jmad_mode_weapon_clas),
    h3_reflexive("mode_ik", jmad_mode_mode_ik),
    BytesRaw("unknown", SIZE=12, VISIBLE=False),
    ENDIAN=">", SIZE=40
    )


jmad_vehicle_suspension = Struct("vehicle_suspension",
    h3_string_id("label"),
    SInt16("graph_index"),
    SInt16("animation"),
    h3_string_id("marker_name"),
    Float("mass_point_offset_0"),
    Float("full_extension_ground_depth"),
    Float("full_compression_ground_depth"),
    h3_string_id("region_name"),
    Float("mass_point_offset_1"),
    Float("expression_ground_depth"),
    Float("compression_ground_depth"),
    ENDIAN=">", SIZE=40
    )


jmad_object_overlay = Struct("object_overlay",
    h3_string_id("label"),
    SInt16("graph_index"),
    SInt16("animation"),
    SInt16("unknown_0", VISIBLE=False),
    SEnum16("function_controls", *jmad_object_overlay_function_controls),
    h3_string_id("function"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=20
    )


jmad_inheritance_list_node_map = Struct("node_map",
    SInt16("local_node"),
    ENDIAN=">", SIZE=2
    )


jmad_inheritance_list_node_map_flag = Struct("node_map_flag",
    Bool32("local_node_flags",
        "node_0_32_64_96_128_160_192_224",
        "node_1_33_65_97_129_161_193_225",
        "node_2_34_66_98_130_162_194_226",
        "node_3_35_67_99_131_163_195_227",
        "node_4_36_68_100_132_164_196_228",
        "node_5_37_69_101_133_165_197_229",
        "node_6_38_70_102_134_166_198_230",
        "node_7_39_71_103_135_167_199_231",
        "node_8_40_72_104_136_168_200_232",
        "node_9_41_73_105_137_169_201_233",
        "node_10_42_74_106_138_170_202_234",
        "node_11_43_75_107_139_171_203_235",
        "node_12_44_76_108_140_172_204_236",
        "node_13_45_77_109_141_173_205_237",
        "node_14_46_78_110_142_174_206_238",
        "node_15_47_79_111_143_175_207_239",
        "node_16_48_80_112_144_176_208_240",
        "node_17_49_81_113_145_177_209_241",
        "node_18_50_82_114_146_178_210_242",
        "node_19_51_83_115_147_179_211_243",
        "node_20_52_84_116_148_180_212_244",
        "node_21_53_85_117_149_181_213_245",
        "node_22_54_86_118_150_182_214_246",
        "node_23_55_87_119_151_183_215_247",
        "node_24_56_88_120_152_184_216_248",
        "node_25_57_89_121_153_185_217_249",
        "node_26_58_90_122_154_186_218_250",
        "node_27_59_91_123_155_187_219_251",
        "node_28_60_92_124_156_188_220_252",
        "node_29_61_93_125_157_189_221_253",
        "node_30_62_94_126_158_190_222_254",
        "node_31_63_95_127_159_191_223_255",
        ),
    ENDIAN=">", SIZE=4
    )


jmad_inheritance_list = Struct("inheritance_list",
    h3_dependency("inherited_graph"),
    h3_reflexive("node_map", jmad_inheritance_list_node_map),
    h3_reflexive("node_map_flags", jmad_inheritance_list_node_map_flag),
    Float("root_z_offset"),
    Bool32("inheritance_flags",
        "tighten_nodes",
        ),
    ENDIAN=">", SIZE=48
    )


jmad_weapon_list = Struct("weapon_list",
    h3_string_id("weapon_name"),
    h3_string_id("weapon_class"),
    ENDIAN=">", SIZE=8
    )


jmad_raw_information_group = Struct("raw_information_group",
    SInt32("member_count"),
    zone_asset("asset_data"),
    ENDIAN=">", SIZE=12
    )


jmad_body = Struct("tagdata",
    h3_dependency("parent_animation_graph"),
    Bool8("inheritance_flags",
        "inherit_root_trans_scale_only",
        "inherit_for_use_on_player",
        ),
    Bool8("private_flags",
        "prepared_for_cache",
        ("imported_with_codec_compressors", 1 << 2),
        ("written_to_cache", 1 << 4),
        "animation_data_reordered",
        ),
    SInt16("animation_codec_pack"),
    h3_reflexive("skeleton_nodes", jmad_skeleton_node),
    h3_reflexive("sound_references", jmad_sound_reference),
    h3_reflexive("effect_references", jmad_effect_reference),
    h3_reflexive("blend_screens", jmad_blend_screen),
    h3_reflexive("legs", jmad_leg),
    h3_reflexive("animations", jmad_animation),
    h3_reflexive("modes", jmad_mode),
    h3_reflexive("vehicle_suspension", jmad_vehicle_suspension),
    h3_reflexive("object_overlays", jmad_object_overlay),
    h3_reflexive("inheritance_list", jmad_inheritance_list),
    h3_reflexive("weapon_list", jmad_weapon_list),
    Bool32("unknown_arm_nodes_1", *("node_%s" % i for i in range(32)), VISIBLE=False),
    Bool32("unknown_arm_nodes_2", *("node_%s" % i for i in range(32, 64)), VISIBLE=False),
    Bool32("unknown_arm_nodes_3", *("node_%s" % i for i in range(64, 96)), VISIBLE=False),
    Bool32("unknown_arm_nodes_4", *("node_%s" % i for i in range(96, 128)), VISIBLE=False),
    Bool32("unknown_arm_nodes_5", *("node_%s" % i for i in range(128, 160)), VISIBLE=False),
    Bool32("unknown_arm_nodes_6", *("node_%s" % i for i in range(160, 192)), VISIBLE=False),
    Bool32("unknown_arm_nodes_7", *("node_%s" % i for i in range(192, 224)), VISIBLE=False),
    Bool32("unknown_arm_nodes_8", *("node_%s" % i for i in range(224, 256)), VISIBLE=False),
    Bool32("unknown_nodes_1", *("node_%s" % i for i in range(32)), VISIBLE=False),
    Bool32("unknown_nodes_2", *("node_%s" % i for i in range(32, 64)), VISIBLE=False),
    Bool32("unknown_nodes_3", *("node_%s" % i for i in range(64, 96)), VISIBLE=False),
    Bool32("unknown_nodes_4", *("node_%s" % i for i in range(96, 128)), VISIBLE=False),
    Bool32("unknown_nodes_5", *("node_%s" % i for i in range(128, 160)), VISIBLE=False),
    Bool32("unknown_nodes_6", *("node_%s" % i for i in range(160, 192)), VISIBLE=False),
    Bool32("unknown_nodes_7", *("node_%s" % i for i in range(192, 224)), VISIBLE=False),
    Bool32("unknown_nodes_8", *("node_%s" % i for i in range(224, 256)), VISIBLE=False),
    h3_rawdata_ref("last_import_results"),
    BytesRaw("unknown", SIZE=12, VISIBLE=False),
    h3_reflexive("raw_information_groups", jmad_raw_information_group),
    ENDIAN=">", SIZE=260
    )


def get():
    return jmad_def

jmad_def = TagDef("jmad",
    h3_blam_header('jmad'),
    jmad_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["jmad"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.hek.defs.objs.tag import HekTag
from reclaimer.util.matrices import euler_2d_to_vector_3d
#from reclaimer.common_descs import valid_objects

class EffeTag(HekTag):

    def calc_internal_data(self):
        HekTag.calc_internal_data(self)

        never_cull = False
        for event in self.data.tagdata.events.STEPTREE:
            for part in event.parts.STEPTREE:
                if part.type.tag_class.enum_name == 'light':
                    never_cull = True

                part.effect_class = part.type.tag_class

                #TODO: There is no good way to do this right now
                #object_types = valid_objects('b').desc[0]['NAME_MAP'].keys()
                #if part.effect_class.enum_name in object_types:
                #    part.effect_class.enum_name = 'object'

            for particle in event.particles.STEPTREE:
                particle.relative_direction_vector[:] = euler_2d_to_vector_3d(
                    *particle.relative_direction
                    )
        self.data.tagdata.flags.never_cull = never_cull



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

def get():
    return matg_def

camera = Struct("camera",
    dependency('camera', "trak"),
    SIZE=16
    )

sound = Struct("sound",
    dependency('sound', "snd!"),
    SIZE=16
    )

look_function = Struct("look_function",
    Float('scale'),
    SIZE=4
    )

player_control = Struct("player_control",
    Float('magnetism_friction'),
    Float('magnetism_adhesion'),
    Float('inconsequential_target_scale'),

    Pad(52),
    float_sec('look_acceleration_time'),
    Float('look_acceleration_scale'),
    float_zero_to_one('look_peg_threshold'),
    float_deg('look_default_pitch_rate'),  # degrees
    float_deg('look_default_yaw_rate'),  # degrees
    Float('look_autolevelling_scale'),

    Pad(20),
    SInt16('minimum_weapon_swap_ticks'),
    SInt16('minimum_autolevelling_ticks'),
    float_rad('minimum_angle_for_vehicle_flip'),  # radians
    reflexive("look_functions", look_function, 16),

    SIZE=128
    )

difficulty_base = QStruct("",
    Float("easy"),
    Float("normal"),
    Float("hard"),
    Float("imposs"),
    ORIENT='h'
    )

difficulty = Struct("difficulty",
    # Health
    Struct("enemy_scales",
        QStruct("damage", INCLUDE=difficulty_base),
        QStruct("vitality", INCLUDE=difficulty_base),
        QStruct("shield", INCLUDE=difficulty_base),
        QStruct("recharge", INCLUDE=difficulty_base),
        ),
    Struct("friend_scales",
        QStruct("damage", INCLUDE=difficulty_base),
        QStruct("vitality", INCLUDE=difficulty_base),
        QStruct("shield", INCLUDE=difficulty_base),
        QStruct("recharge", INCLUDE=difficulty_base),
        ),
    QStruct("infection_form_vitality_scales", INCLUDE=difficulty_base),

    Pad(16),
    # Enemy ranged fire
    Struct("ranged_combat_scales",
        QStruct("rate_of_fire", INCLUDE=difficulty_base),
        QStruct("projectile_error", INCLUDE=difficulty_base),
        QStruct("burst_error", INCLUDE=difficulty_base),
        QStruct("new_target_delay", INCLUDE=difficulty_base),
        QStruct("burst_separation", INCLUDE=difficulty_base),
        QStruct("target_tracking", INCLUDE=difficulty_base),
        QStruct("target_leading", INCLUDE=difficulty_base),
        QStruct("overcharge_chance", INCLUDE=difficulty_base),
        QStruct("special_fire_delay", INCLUDE=difficulty_base),
        QStruct("guidance_vs_player", INCLUDE=difficulty_base),
        ),

    Struct("close_combat_scales",
        QStruct("melee_delay_base",
            GUI_NAME="melee delay base(not a scale)", INCLUDE=difficulty_base
            ),
        QStruct("melee_delay", INCLUDE=difficulty_base),

        Pad(16),
        # Grenades
        QStruct("grenade_chance", INCLUDE=difficulty_base),
        QStruct("grenade_timer", INCLUDE=difficulty_base),
        ),

    Pad(48),
    # Placement
    Struct("major_upgrade_fractions",
        QStruct("normal", INCLUDE=difficulty_base),
        QStruct("few", INCLUDE=difficulty_base),
        QStruct("many", INCLUDE=difficulty_base),
        ),

    SIZE=644
    )

grenade = Struct("grenade",
    SInt16('maximum_count'),
    SInt16('mp_spawn_default'),
    dependency('throwing_effect', "effe"),
    dependency('hud_interface', "grhi"),
    dependency('equipment', "eqip"),
    dependency('projectile', "proj"),
    SIZE=68
    )

rasterizer_data = Struct("rasterizer_data",
    Struct("function_textures",
        dependency('distance_attenuation', "bitm"),
        dependency('vector_normalization', "bitm"),
        dependency('atmospheric_fog_density', "bitm"),
        dependency('planar_fog_density', "bitm"),
        dependency('linear_corner_fade', "bitm"),
        dependency('active_camouflage_distortion', "bitm"),
        dependency('glow', "bitm"),
        Pad(60),
        ),

    # Default textures
    Struct("default_textures",
        dependency('default_2d', "bitm"),
        dependency('default_3d', "bitm"),
        dependency('default_cubemap', "bitm"),
        ),

    # Experimental textures
    Struct("experimental_textures",
        dependency('test0', "bitm"),
        dependency('test1', "bitm"),
        dependency('test2', "bitm"),
        dependency('test3', "bitm"),
        ),

    # video effect textures
    Struct("video_effect_textures",
        dependency('video_scanline_map', "bitm"),
        dependency('video_noise_map', "bitm"),
        Pad(52),
        ),

    # Active camouflage
    Struct("active_camouflage",
        Bool16("flags",
            "tint_edge_density"
            ),
        Pad(2),
        Float('refraction_amount', SIDETIP="pixels"),  # pixels
        Float('distance_falloff'),
        QStruct('tint_color', INCLUDE=rgb_float),
        Float('hyper_stealth_refraction_amount', SIDETIP="pixels"),  # pixels
        Float('hyper_stealth_distance_falloff'),
        QStruct('hyper_stealth_tint_color', INCLUDE=rgb_float),
        ),

    # PC textures
    dependency('distance_attenuation_2d', "bitm"),

    SIZE=428
    )

interface_bitmaps = Struct("interface_bitmaps",
    dependency('font_system', "font"),
    dependency('font_terminal', "font"),
    dependency('screen_color_table', "colo"),
    dependency('hud_color_table', "colo"),
    dependency('editor_color_table', "colo"),
    dependency('dialog_color_table', "colo"),
    dependency('hud_globals', "hudg"),
    dependency('motion_sensor_sweep_bitmap', "bitm"),
    dependency('motion_sensor_sweep_bitmap_mask', "bitm"),
    dependency('multiplayer_hud_bitmap', "bitm"),
    dependency('localization', "str#"),
    dependency('hud_digits_definition', "hud#"),
    dependency('motion_sensor_blip', "bitm"),
    dependency('interface_goo_map1', "bitm"),
    dependency('interface_goo_map2', "bitm"),
    dependency('interface_goo_map3', "bitm"),
    SIZE=304
    )

cheat_weapon = Struct("weapon",
    dependency('weapon', valid_items),
    SIZE=16
    )

cheat_powerup = Struct("powerup",
    dependency('powerup', "eqip"),
    SIZE=16
    )

vehicle = Struct("vehicle",
    dependency('vehicle', "vehi"),
    SIZE=16
    )

multiplayer_information = Struct("multiplayer_information",
    dependency('flag', valid_items),
    dependency('unit', valid_units),
    reflexive('vehicles', vehicle, 20, DYN_NAME_PATH='.vehicle.filepath'),
    dependency('hill_shader', valid_shaders),
    dependency('flag_shader', valid_shaders),
    dependency('ball', valid_items),
    reflexive('sounds', sound, 60, DYN_NAME_PATH='.sound.filepath'),
    SIZE=160
    )

player_information = Struct("player_information",
    dependency('unit', valid_units),

    Pad(28),
    float_wu_sec("walking_speed"),  # world units/second
    Float("double_speed_multiplier", SIDETIP="[1.0,+inf]", MIN=1.0),
    float_wu_sec("run_forward"),  # world units/second
    float_wu_sec("run_backward"),  # world units/second
    float_wu_sec("run_sideways"),  # world units/second
    float_wu_sec_sq("run_acceleration",
                    UNIT_SCALE=per_sec_unit_scale),  # world units/second^2
    float_wu_sec("sneak_forward"),  # world units/second
    float_wu_sec("sneak_backward"),  # world units/second
    float_wu_sec("sneak_sideways"),  # world units/second
    float_wu_sec_sq("sneak_acceleration",
                    UNIT_SCALE=per_sec_unit_scale),  # world units/second^2
    float_wu_sec_sq("airborne_acceleration",
                    UNIT_SCALE=per_sec_unit_scale),  # world units/second^2
    Float("speed_multiplier", SIDETIP="multiplayer only"),  # multiplayer only

    Pad(12),
    QStruct("grenade_origin", INCLUDE=xyz_float),

    Pad(12),
    float_zero_to_one("stun_movement_penalty"),
    float_zero_to_one("stun_turning_penalty"),
    float_zero_to_one("stun_jumping_penalty"),
    Float("minimum_stun_time"),
    Float("maximum_stun_time"),

    Pad(8),
    from_to_sec("first_person_idle_time"),
    float_zero_to_one("first_person_skip_fraction"),

    Pad(16),
    dependency('coop_respawn_effect', "effe"),
    SIZE=244
    )

first_person_interface = Struct("first_person_interface",
    dependency('first_person_hands', valid_models),
    dependency('base_bitmap', "bitm"),
    dependency('shield_meter', "metr"),
    QStruct('shield_meter_origin',
        SInt16('x'), SInt16('y'), ORIENT='h'
        ),
    dependency('body_meter', "metr"),
    QStruct('body_meter_origin',
        SInt16('x'), SInt16('y'), ORIENT='h'
        ),
    dependency('night_vision_toggle_on_effect', "effe"),
    dependency('night_vision_toggle_off_effect', "effe"),

    SIZE=192
    )

falling_damage = Struct("falling_damage",
    Pad(8),
    QStruct("harmful_falling_distance", INCLUDE=from_to),
    dependency('falling_damage', "jpt!"),

    Pad(8),
    Float("maximum_falling_distance"),
    dependency('distance_damage', "jpt!"),
    dependency('vehicle_environment_collision_damage', "jpt!"),
    dependency('vehicle_killed_unit_damage', "jpt!"),
    dependency('vehicle_collision_damage', "jpt!"),
    dependency('flaming_death_damage', "jpt!"),
    SIZE=152
    )

particle_effect = Struct("particle_effect",
    dependency('particle_type', "part"),
    Bool32("flags", *blend_flags),
    float_wu("density"),
    QStruct("velocity_scale", INCLUDE=from_to),

    Pad(4),
    from_to_rad_sec("angular_velocity"),  # radians/second

    Pad(8),
    from_to_wu("radius"),  # world units

    Pad(8),
    QStruct("tint_lower_bound", INCLUDE=argb_float),
    QStruct("tint_upper_bound", INCLUDE=argb_float),
    SIZE=128
    )

material = Struct("material",
    Pad(148),
    # Vehicle terrain parameters
    Float("ground_friction_scale"),
    Float("ground_friction_normal_k1_scale"),
    Float("ground_friction_normal_k0_scale"),
    Float("ground_depth_scale"),
    Float("ground_damp_fraction_scale"),

    # Breakable surface parameters
    Pad(556),
    Float("maximum_vitality"),

    Pad(12),
    dependency('effect', "effe"),
    dependency('sound', "snd!"),

    Pad(24),
    reflexive("particle_effects", particle_effect, 8,
        DYN_NAME_PATH='.particle_type.filepath'),

    Pad(60),
    dependency('melee_hit_sound', "snd!"),
    SIZE=884
    )

playlist_member = Struct("playlist_member",
    ascii_str32("map_name"),
    ascii_str32("game_variant"),
    SInt32('minimum_experience'),
    SInt32('maximum_experience'),
    SInt32('minimum_player_count'),
    SInt32('maximum_player_count'),
    SInt32('rating'),
    SIZE=148
    )

matg_body = Struct('tagdata',
    Pad(248),
    reflexive("sounds", sound, 2,
        "enter water", "exit water"),
    reflexive("cameras", camera, 1),
    reflexive("player_controls", player_control, 1),
    reflexive("difficulties", difficulty, 1),
    reflexive("grenades", grenade, 2, *grenade_types),
    reflexive("rasterizer_datas", rasterizer_data, 1),
    reflexive("interface_bitmaps", interface_bitmaps, 1),
    reflexive("cheat_weapons", cheat_weapon, 20,
        DYN_NAME_PATH='.weapon.filepath'),
    reflexive("cheat_powerups", cheat_powerup, 20,
        DYN_NAME_PATH='.powerup.filepath'),
    reflexive("multiplayer_informations", multiplayer_information, 1),
    reflexive("player_informations", player_information, 1),
    reflexive("first_person_interfaces", first_person_interface, 1),
    reflexive("falling_damages", falling_damage, 1),
    reflexive("materials", material, len(materials_list), *materials_list),
    reflexive("playlist_members", playlist_member, 20,
        DYN_NAME_PATH='.map_name'),

    SIZE=428
    )

matg_def = TagDef("matg",
    blam_header('matg', 3),
    matg_body,

    ext=".globals", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


sfx__sound_effect_filter = Struct("filter",
    SInt32("filter_type"),
    SInt32("filter_width"),
    Pad(64),
    ENDIAN=">", SIZE=72
    )


sfx__sound_effect_pitch_lfo = Struct("pitch_lfo",
    Pad(48),
    ENDIAN=">", SIZE=48
    )


sfx__sound_effect_filter_lfo = Struct("filter_lfo",
    Pad(64),
    ENDIAN=">", SIZE=48
    )


sfx__sound_effect_sound_effect_component = Struct("component",
    h3_dependency("sound"),
    Pad(4),
    SInt32("flags"),
    ENDIAN=">", SIZE=24
    )


sfx__sound_effect_sound_effect_template_collection_parameter = Struct("parameter",
    h3_string_id("name"),
    Pad(20),
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=44
    )


sfx__sound_effect_sound_effect_template_collection = Struct("template_collection",
    h3_string_id("dsp_effect"),
    h3_reflexive("parameters", sfx__sound_effect_sound_effect_template_collection_parameter),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


sfx__sound_effect_sound_effect = Struct("sound_effect",
    h3_dependency("unknown_0"),
    h3_reflexive("components", sfx__sound_effect_sound_effect_component),
    h3_reflexive("template_collection", sfx__sound_effect_sound_effect_template_collection),
    BytesRaw("unknown_1", SIZE=32, VISIBLE=False),
    ENDIAN=">", SIZE=72
    )


sfx__sound_effect = Struct("sound_effect",
    h3_string_id("name"),
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    Bool32("flags",
        "use_3d_radio_hack",
        ),
    BytesRaw("unknown_1", SIZE=8, VISIBLE=False),
    h3_reflexive("filter", sfx__sound_effect_filter),
    h3_reflexive("pitch_lfo", sfx__sound_effect_pitch_lfo),
    h3_reflexive("filter_lfo", sfx__sound_effect_filter_lfo),
    h3_reflexive("sound_effect", sfx__sound_effect_sound_effect),
    ENDIAN=">", SIZE=76
    )


sfx__body = Struct("tagdata",
    h3_reflexive("sound_effects", sfx__sound_effect),
    ENDIAN=">", SIZE=12
    )


def get():
    return sfx__def

sfx__def = TagDef("sfx+",
    h3_blam_header('sfx+'),
    sfx__body,

    ext=".%s" % h3_tag_class_fcc_to_ext["sfx+"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...os_v3_hek.defs.cont import *

cont_body = dict(cont_body)
cont_body[4] = reflexive(
    "shader_extensions",
    Struct("shader_extension", INCLUDE=os_shader_extension),
    1)


def get():
    return cont_def

cont_def = TagDef("cont",
    blam_header("cont", 3),
    cont_body,

    ext=".contrail", endian=">", tag_cls=HekTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DeadCanadian
# 	named some tags
# revision: 3		author: Lord Zedd
# 	Updated and copypasted.
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


lst3_list_widget_item = Struct("list_widget_item",
    Bool32("flags", *unknown_flags_32),
    h3_string_id("name"),
    SInt16("unknown", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_offset"),
    SInt16("widescreen_x_offset"),
    SInt16("widescreen_y_unknown"),
    SInt16("widescreen_x_unknown"),
    SInt16("standard_y_offset"),
    SInt16("standard_x_offset"),
    SInt16("standard_y_unknown"),
    SInt16("standard_x_unknown"),
    h3_dependency("animation"),
    h3_string_id("target"),
    ENDIAN=">", SIZE=48
    )


lst3_body = Struct("tagdata",
    Bool32("flags",
        ("horizontal", 1 << 4),
        "loops",
        ),
    h3_string_id("name"),
    SInt16("unknown", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_offset"),
    SInt16("widescreen_x_offset"),
    SInt16("widescreen_y_unknown"),
    SInt16("widescreen_x_unknown"),
    SInt16("standard_y_offset"),
    SInt16("standard_x_offset"),
    SInt16("standard_y_unknown"),
    SInt16("standard_x_unknown"),
    h3_dependency("animation"),
    h3_string_id("data_source_name"),
    h3_dependency("skin"),
    SInt32("row_count"),
    h3_reflexive("list_widget_items", lst3_list_widget_item),
    h3_dependency("up_arrow_bitmap"),
    h3_dependency("down_arrow_bitmap"),
    ENDIAN=">", SIZE=112
    )


def get():
    return lst3_def

lst3_def = TagDef("lst3",
    h3_blam_header('lst3'),
    lst3_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["lst3"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Veegie
# 	Fixed some shit
# revision: 3		author: Lord Zedd
# 	Standardizing
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


sefc_screen_effect = Struct("screen_effect",
    h3_string_id("name"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    h3_rawdata_ref("function_0", VISIBLE=False),
    Float("duration"),
    h3_rawdata_ref("function_1", VISIBLE=False),
    h3_rawdata_ref("function_2", VISIBLE=False),
    Float("light_intensity"),
    Pad(8),
    Float("saturation"),
    Float("color_muting"),
    Float("brightness"),
    Float("darkness"),
    Float("shadow_brightness"),
    color_rgb_float("tint"),
    color_rgb_float("tone"),
    ENDIAN=">", SIZE=132
    )


sefc_body = Struct("tagdata",
    h3_reflexive("screen_effect", sefc_screen_effect),
    ENDIAN=">", SIZE=12
    )


def get():
    return sefc_def

sefc_def = TagDef("sefc",
    h3_blam_header('sefc'),
    sefc_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["sefc"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

animation_comment = """
Played immediately after the old unit's transform out animation.
The new actor is braindead during the animation, and is invincible until it ends.
"""

vitality_inheritance_overrides = (
    'neither',
    'both',
    'shield_only',
    'health_only'
    )

keyframe_action = Struct("keyframe_action",
    SEnum16("keyframe",
        "primary",
        "secondary",
        "final",
        ),
    SEnum16("rider_handling",
        'none',
        'kill',
        'eject',
        ),
    SEnum16("target",
        'self',
        'riders',
        ),
    Pad(2),
    dependency_os("damage_effect", "jpt!"),
    dependency_os("effect", "effe"),
    ascii_str32("effect_marker"),
    SIZE=72
    )

transform_in_target = Struct("transform_in_target",
    ascii_str32("target_name"),
    Bool16("flags",
        'try_to_use_existing_unit',
        'drop_weapon',
        'inherit_seated_units',
        'delete_attached_actors',
        ),

    Pad(2),
    QStruct("selection_chances",
        Float("easy"), Float("normal"), Float("hard"), Float("imposs"),
        ORIENT="h"
        ),

    Struct("ai",
        dependency_os("actor_variant", "actv"),
        SEnum16("encounter_squad_handling",
            'inherit_from_old_unit',
            'inherit_from_attacker',
            'free_actor',
            ),
        Pad(2),
        SEnum16("team_handling",
            'inherit_from_old_unit',
            'inherit_from_attacker',
            'override',
            ),
        SEnum16("team_override", *unit_teams),
        SEnum16("initial_state_handling",
            'inherit',
            'override',
            ),
        SEnum16("initial_state_override", *actor_states),
        SEnum16("return_state_handling",
            'inherit',
            'override',
            'actor_default',
            ),
        SEnum16("return_state_override", *actor_states),
        Pad(4),
        ),

    Struct("animation",
        ascii_str32("transform_in_anim"),
        reflexive("keyframe_actions", keyframe_action, 9,
            DYN_NAME_PATH='.effect_marker'),
        COMMENT=animation_comment
        ),

    Struct("vitality",
        Pad(4),
        SEnum16("inheritance", *vitality_inheritance_overrides),
        SEnum16("override", *vitality_inheritance_overrides),
        Float("shield_override"),
        Float("health_override"),
        ),
    SIZE=172
    )

avti_body = Struct("tagdata",
    reflexive("targets", transform_in_target, 16,
        DYN_NAME_PATH='.target_name'),
    SIZE=36
    )

def get():
    return avti_def

avti_def = TagDef("avti",
    blam_header_os('avti', 1),
    avti_body,

    ext=".actor_variant_transform_in", endian=">", tag_cls=HekTag
    )
